from pathlib import Path
from typing import Callable, Dict, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from ._base_inferer import BaseInferer


class ResizeInferer(BaseInferer):
    def __init__(
        self,
        model: nn.Module,
        input_path: Union[Path, str],
        out_activations: Dict[str, str],
        out_boundary_weights: Dict[str, bool],
        resize: Tuple[int, int],
        instance_postproc: str,
        padding: int = None,
        batch_size: int = 8,
        normalization: str = None,
        device: str = "cuda",
        n_devices: int = 1,
        save_intermediate: bool = False,
        save_dir: Union[Path, str] = None,
        save_format: str = ".mat",
        checkpoint_path: Union[Path, str] = None,
        n_images: int = None,
        type_post_proc: Callable = None,
        sem_post_proc: Callable = None,
        **kwargs,
    ) -> None:
        """Resize inference for a folder of images.

        Resizes the image before inputting to the model.

        NOTE: This class assumes that all the images in the input folder
        have the same shape. If the images have different shapes,
        `batch_size` needs to be set to 1 for this to work.

        Parameters
        ----------
            model : nn.Module
                A segmentation model.
            input_path : Path | str
                Path to a folder of images or to hdf5 db.
            out_activations : Dict[str, str]
                Dictionary of head names mapped to a string value that specifies the
                activation applied at the head. E.g. {"type": "tanh", "cellpose": None}
                Allowed values: "softmax", "sigmoid", "tanh", None.
            out_boundary_weights : Dict[str, bool]
                Dictionary of head names mapped to a boolean value. If the value is
                True, after a prediction, a weight matrix is applied that assigns bigger
                weight on pixels in the center and less weight to pixels on the tile
                boundaries. helps dealing with prediction artefacts on the boundaries.
                E.g. {"type": False, "cellpose": True}
            resize : Tuple[int, int]:
                The resized size for the input patches that are fed to the segmentation
                model.
            instance_postproc : str
                The post-processing method for the instance segmentation mask. One of:
                "cellpose", "omnipose", "stardist", "hovernet", "dcan", "drfns", "dran"
            padding : int, optional
                The amount of reflection padding for the input images.
            batch_size : int, default=8
                Number of images loaded from the folder at every batch.
            normalization : str, optional
                Apply img normalization (Same as during training). One of "dataset",
                "minmax", "norm", "percentile", None.
            device : str, default="cuda"
                The device of the input and model. One of: "cuda", "cpu"
            n_devices : int, default=1
                Number of devices (cpus/gpus) used for inference.
                The model will be copied into these devices.
            save_intermediate : bool, default=False
                If True, intermediate soft masks will be saved into `soft_masks` var.
            save_intermediate : bool, default=False
                If True, intermediate soft masks will be saved into `soft_masks` var.
            save_format : str, default=".mat"
                The file format for the saved output masks. One of (".mat", ".json").
                The ".json" option will save masks into geojson format.
            checkpoint_path : Path | str, optional
                Path to the model weight checkpoints.
            n_images : int, optional
                First n-number of images used from the `input_path`.
            type_post_proc : Callable, optional
                A post-processing function for the type maps. If not None, overrides
                the default.
            sem_post_proc : Callable, optional
                A post-processing function for the semantc seg maps. If not None,
                overrides the default.
            **kwargs:
                Arbitrary keyword arguments expecially for post-processing and saving.

        Examples
        --------
            >>> # initialize model and paths
            >>> model = cellpose_base(len(type_classes))
            >>> inputs = "/path/to/imgs"
            >>> ckpt_path = "/path/to/myweights.ckpt"

            >>> # initialize output head args
            >>> out_activations={"type": "softmax", "cellpose": None}
            >>> out_boundary_weights={"type": None, "cellpose": None}

            >>> inferer = ResizeInferer(
                    model=model,
                    input_path=inputs,
                    checkpoint_path=ckpt_path,
                    out_activations=out_activations,
                    out_boundary_weights=out_boundary_weights,
                    resize=(256, 256),
                    instance_postproc="cellpose",
                    padding=0,
                    normalization="minmax" # This needs to be same as during training
                )
            >>> inferer.infer()
        """
        super().__init__(
            model=model,
            input_path=input_path,
            out_activations=out_activations,
            out_boundary_weights=out_boundary_weights,
            patch_size=resize,
            padding=padding,
            batch_size=batch_size,
            normalization=normalization,
            instance_postproc=instance_postproc,
            device=device,
            n_devices=n_devices,
            save_intermediate=save_intermediate,
            save_dir=save_dir,
            save_format=save_format,
            checkpoint_path=checkpoint_path,
            n_images=n_images,
            type_post_proc=type_post_proc,
            sem_post_proc=sem_post_proc,
            **kwargs,
        )

    def _infer_batch(self, input_batch: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Infer one batch of images."""
        inp_shape = tuple(input_batch.shape[2:])

        if self.padding:
            pad = self.padding // 2
            input_batch = F.pad(
                input_batch.float(), pad=(pad, pad, pad, pad), mode="reflect"
            )
            inp_shape = tuple(input_batch.shape[2:])

        input_batch = F.interpolate(input_batch, self.patch_size)
        batch = input_batch.to(self.device).float()
        logits = self.predictor.forward_pass(batch)

        probs = {}
        for k, logit in logits.items():
            prob = self.predictor.classify(logit, **self.head_kwargs[k])

            if self.padding:
                probs[k] = F.interpolate(prob, inp_shape)[..., pad:-pad, pad:-pad]
            else:
                probs[k] = F.interpolate(prob, inp_shape)

        return probs



import pytest
import torch
import torch.nn as nn

from cellseg_models_pytorch.modules.misc_modules import (
    ChannelPool,
    StyleBlock,
    StyleReshape,
)


@pytest.mark.parametrize("in_channels", [32, 16])
@pytest.mark.parametrize("out_channels", [16, 32])
def test_chpool_fwdbwd(in_channels, out_channels):
    x = torch.rand([1, in_channels, 16, 16])
    chpool = ChannelPool(in_channels, out_channels)
    out = chpool(x)

    out.mean().backward()

    assert out.shape[1] == out_channels


@pytest.mark.parametrize("in_channels", [32])
@pytest.mark.parametrize("style_channels", [32, 16])
@pytest.mark.parametrize("out_channels", [16, 32])
def test_stylefwdbwd(in_channels, style_channels, out_channels):
    x = torch.rand([1, in_channels, 16, 16])
    make_style = StyleReshape(in_channels, style_channels)
    style_block = StyleBlock(style_channels, out_channels)
    conv = nn.Conv2d(in_channels, out_channels, 3, padding=1)

    style = make_style(x)
    x = conv(x)
    out = style_block(x, style)

    out.mean().backward()

    assert out.shape[1] == out_channels



import pytest
import torch

from cellseg_models_pytorch.models import (
    MultiTaskUnet,
    cellpose_base,
    cellpose_plus,
    hovernet_base,
    hovernet_plus,
    hovernet_small,
    hovernet_small_plus,
    omnipose_base,
    omnipose_plus,
    stardist_base,
    stardist_base_multiclass,
    stardist_plus,
)


@pytest.mark.parametrize(
    "model", [hovernet_base, hovernet_plus, hovernet_small_plus, hovernet_small]
)
@pytest.mark.parametrize("style_channels", [None, 32])
def test_hovernet_fwdbwd(model, style_channels):
    x = torch.rand([1, 3, 64, 64])
    m = model(
        type_classes=3,
        sem_classes=3,
        style_channels=style_channels,
    )
    y = m(x)
    y["hovernet"].mean().backward()

    assert y["type"].shape == x.shape


@pytest.mark.parametrize(
    "model", [stardist_base, stardist_plus, stardist_base_multiclass]
)
@pytest.mark.parametrize("style_channels", [None, 32])
def test_stardist_fwdbwd(model, style_channels):
    n_rays = 3
    x = torch.rand([1, 3, 64, 64])
    m = model(
        n_rays=n_rays,
        type_classes=3,
        sem_classes=3,
        style_channels=style_channels,
    )
    y = m(x)
    y["stardist"].mean().backward()

    assert y["stardist"].shape == torch.Size([1, n_rays, 64, 64])


@pytest.mark.parametrize("model", [cellpose_base, cellpose_plus])
def test_cellpose_fwdbwd(model):
    x = torch.rand([1, 3, 64, 64])
    m = model(type_classes=3, sem_classes=3)
    y = m(x)
    y["cellpose"].mean().backward()

    if "sem" in y.keys():
        assert y["sem"].shape == torch.Size([1, 3, 64, 64])

    assert y["cellpose"].shape == torch.Size([1, 2, 64, 64])


@pytest.mark.parametrize("model", [omnipose_base, omnipose_plus])
def test_omnipose_fwdbwd(model):
    x = torch.rand([1, 3, 64, 64])
    m = model(type_classes=3, sem_classes=3)
    y = m(x)
    y["omnipose"].mean().backward()

    if "sem" in y.keys():
        assert y["sem"].shape == torch.Size([1, 3, 64, 64])

    assert y["omnipose"].shape == torch.Size([1, 2, 64, 64])


def test_multitaskunet_fwdbwd():
    x = torch.rand([1, 3, 64, 64])
    m = MultiTaskUnet(
        decoders=("sem",),
        heads={"sem": {"sem": 3}},
        n_conv_layers={"sem": (1, 1, 1, 1)},
        n_conv_blocks={"sem": ((2,), (2,), (2,), (2,))},
        out_channels={"sem": (128, 64, 32, 16)},
        long_skips={"sem": "unet"},
        dec_params={"sem": None},
    )
    y = m(x)
    y["sem"].mean().backward()

    assert y["sem"].shape == torch.Size([1, 3, 64, 64])



import torch
import torch.nn as nn

from .base_modules import Conv, Norm

__all__ = ["ChannelPool", "StyleBlock", "StyleReshape", "LayerScale"]


class LayerScale(nn.Module):
    def __init__(
        self, dim: int, init_values: float = 1e-5, inplace: bool = False
    ) -> None:
        """Learnable scaling factor for transformer components.

        I.e. this is used to scale the attention matrix and the mlp head
        of the transformer.

        NOTE: Copied from timm vision_transformer.py


        Parameters
        ----------
            dim : int
                The dimensionality of the input.
            init_values : float, default=1e-5
                Initialization values for the learnable weights.
            inplace : bool, default=False
                Flag, whether the scaling is an inplace operation.
        """
        super().__init__()
        self.dim = dim
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the layer scaling."""
        return x.mul_(self.gamma) if self.inplace else x * self.gamma

    def extra_repr(self) -> str:
        """Add extra to repr."""
        return f"dim={self.dim}, inplace={self.inplace}"


class ChannelPool(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        normalization: str = "bn",
        convolution: str = "conv",
        **kwargs,
    ) -> None:
        """Channel pooling/downsampling module for convenience.

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", "lrn", None
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.proj = nn.Sequential(
            Conv(
                convolution,
                in_channels=in_channels,
                bias=False,
                out_channels=out_channels,
                kernel_size=1,
                padding=0,
            ),
            Norm(normalization, num_features=self.out_channels),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the channel pool."""
        return self.proj(x)


class StyleBlock(nn.Module):
    def __init__(self, style_channels: int, out_channels: int) -> None:
        """Add a style vector to the input tensor. See Cellpose.

        Cellpose:
        - https://www.nature.com/articles/s41592-020-01018-x

        Parameters
        ----------
            style_channels : int
                Number of style vector channels.
            out_channels : int
                Number of output channels.
        """
        super().__init__()
        self.out_channels = out_channels
        self.full = nn.Linear(style_channels, out_channels)

    def forward(self, x: torch.Tensor, style_feat: torch.Tensor) -> torch.Tensor:
        """Forward style block."""
        style_feat = self.full(style_feat)
        out = x + style_feat.unsqueeze(-1).unsqueeze(-1)

        return out


class StyleReshape(nn.Module):
    def __init__(
        self,
        in_channels: int,
        style_channels: int,
        normalization: str = "bn",
        convolution: str = "conv",
        **kwargs,
    ) -> None:
        """Reshape feature-map into a style-vector. See Cellpose.

        Cellpose:
        - https://www.nature.com/articles/s41592-020-01018-x

        Takes in a feature map (B, C, H, W). Then averages, and normalizes it
        into to a style feature vector.

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            style_channels : int
                Number of style vector channels.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", "lrn", None
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"

        Returns
        -------
            torch.Tensor:
                Style vector. Shape: (B, C).
        """
        super().__init__()
        self.flatten = nn.Flatten()
        self.out_channels = style_channels

        self.downsample = None
        if in_channels != style_channels:
            self.downsample = ChannelPool(
                in_channels=in_channels,
                out_channels=style_channels,
                convolution=convolution,
                normalization=normalization,
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward style-block."""
        if self.downsample is not None:
            x = self.downsample(x)

        style = x.mean((2, 3), keepdim=True)
        style = self.flatten(style)
        style = style / torch.sum(style**2, axis=1, keepdim=True) ** 0.5

        return style



from typing import Any, Dict

import torch
import torch.nn as nn

from .base_modules import Activation, Norm

__all__ = ["Mlp", "ConvMlp", "MlpBlock"]


class Mlp(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mlp_ratio: int = 2,
        activation: str = "star_relu",
        dropout: float = 0.0,
        bias: bool = False,
        out_channels: int = None,
        act_kwargs: Dict[str, Any] = None,
        **kwargs,
    ) -> None:
        """MLP token mixer.

        - MetaFormer: https://arxiv.org/abs/2210.13452
        - MLP-Mixer: https://arxiv.org/abs/2105.01601

        - Input shape: (B, N, embed_dim)
        - Output shape: (B, seq_len, embed_dim)

        Parameters
        ----------
            in_channels : int
                Number of input features.
            mlp_ratio : int, default=2
                Scaling factor to get the number hidden features from the `in_channels`.
            activation : str, default="star_relu"
                The name of the activation function.
            dropout : float, default=0.0
                Dropout ratio.
            bias : bool, default=False
                Flag whether to use bias terms in the nn.Linear modules.
            out_channels : int, optional
                Number of out channels. If None `out_channels = in_channels`
            act_kwargs : Dict[str, Any], optional
                Arbitrary key-word arguments for the activation function.
        """
        super().__init__()
        act_kwargs = act_kwargs if act_kwargs is not None else {}
        self.out_channels = in_channels if out_channels is None else out_channels
        hidden_channels = int(mlp_ratio * in_channels)

        self.fc1 = nn.Linear(in_channels, hidden_channels, bias=bias)
        self.act = Activation(activation, **act_kwargs)
        self.drop1 = nn.Dropout(dropout)
        self.fc2 = nn.Linear(hidden_channels, self.out_channels, bias=bias)
        self.drop2 = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass of the MLP token mixer."""
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)

        return x


class ConvMlp(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mlp_ratio: int = 2,
        activation: str = "star_relu",
        dropout: float = 0.0,
        bias: bool = False,
        out_channels: int = None,
        act_kwargs: Dict[str, Any] = None,
        **kwargs,
    ) -> None:
        """Mlp layer implemented with dws convolution.

        Input shape: (B, in_channels, H, W).
        Output shape: (B, out_channels, H, W).

        Parameters
        ----------
            in_channels : int
                Number of input features.
            mlp_ratio : int, default=2
                Scaling factor to get the number hidden features from the `in_channels`.
            activation : str, default="star_relu"
                The name of the activation function.
            dropout : float, default=0.0
                Dropout ratio.
            bias : bool, default=False
                Flag whether to use bias terms in the nn.Linear modules.
            out_channels : int, optional
                Number of out channels. If None `out_channels = in_channels`
            act_kwargs : Dict[str, Any], optional
                Arbitrary key-word arguments for the activation function.
        """
        super().__init__()
        act_kwargs = act_kwargs if act_kwargs is not None else {}
        self.out_channels = in_channels if out_channels is None else out_channels
        self.hidden_channels = int(mlp_ratio * in_channels)
        self.fc1 = nn.Conv2d(in_channels, self.hidden_channels, 1, bias=bias)
        self.dwconv = nn.Conv2d(
            in_channels, in_channels, 3, 1, 1, bias=bias, groups=in_channels
        )
        self.act = Activation(activation, **act_kwargs)
        self.fc2 = nn.Conv2d(self.hidden_channels, self.out_channels, 1, bias=bias)
        self.drop = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of conv-mlp."""
        x = self.fc1(x)

        x = self.dwconv(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class MlpBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mlp_type: str = "linear",
        mlp_ratio: int = 2,
        activation: str = "star_relu",
        act_kwargs: Dict[str, Any] = None,
        dropout: float = 0.0,
        bias: bool = False,
        normalization: str = "ln",
        norm_kwargs: Dict[str, Any] = None,
    ) -> None:
        """Residual Mlp block.

        I.e. norm -> mlp -> residual

        Parameters
        ----------
            in_channels : int
                Number of input features.
            mlp_type : str, default="linear"
                Flag for either nn.Linear or nn.Conv2d mlp-layer.
                One of "conv", "linear".
            mlp_ratio : int, default=2
                Scaling factor to get the number hidden features from the `in_channels`.
            activation : str, default="star_relu"
                The name of the activation function.
            act_kwargs : Dict[str, Any], optional
                key-word args for the activation module.
            dropout : float, default=0.0
                Dropout ratio.
            bias : bool, default=False
                Flag whether to use bias terms in the nn.Linear modules.
            normalization : str, default="ln"
                The name of the normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", "lrn", None
            norm_kwargs : Dict[str, Any], optional
                key-word args for the normalization layer. Ignored if normalization
                is None.
        """
        super().__init__()
        allowed = ("conv", "linear")
        if mlp_type not in allowed:
            raise ValueError(
                f"Illegal `mlp_type` given. Got: {mlp_type}. Allowed: {allowed}."
            )

        norm_kwargs = norm_kwargs if norm_kwargs is not None else {}
        act_kwargs = act_kwargs if act_kwargs is not None else {}
        self.norm = Norm(normalization, **norm_kwargs)
        MlpHead = Mlp if mlp_type == "linear" else ConvMlp

        self.mlp = MlpHead(
            in_channels=in_channels,
            mlp_ratio=mlp_ratio,
            activation=activation,
            dropout=dropout,
            bias=bias,
            act_kwargs=act_kwargs,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the Metaformer Mlp-block."""
        residual = x

        x = self.norm(x)
        x = self.mlp(x)

        return x + residual



from typing import Any, Dict

import torch
import torch.nn as nn

from .attention_modules import MSCA
from .base_modules import Activation, Conv, Identity, Norm
from .mlp import Mlp
from .self_attention_modules import SelfAttention

__all__ = ["MSCAN", "Pooling", "TokenMixer", "TokenMixerBlock"]


class MSCAN(nn.Module):
    def __init__(
        self, in_channels: int, conv: str = "conv", activation: str = "gelu", **kwargs
    ) -> None:
        """Create MSCAN spatial attention module.

        Parameters
        ----------
            conv : str, default="conv"
                Convolution layer type.
            activation : str, default="relu"
                Activation layer after squeeze.
        """
        super().__init__()
        self.in_channels = in_channels
        self.proj_1 = Conv(
            conv, in_channels=in_channels, out_channels=in_channels, kernel_size=1
        )
        self.activation = Activation(activation)
        self.spatial_gating_unit = MSCA(in_channels)
        self.proj_2 = Conv(
            conv, in_channels=in_channels, out_channels=in_channels, kernel_size=1
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of MSCAN attention."""
        residual = x

        x = self.proj_1(x)
        x = self.activation(x)
        x = self.spatial_gating_unit(x)
        x = self.proj_2(x)

        return x + residual


class Pooling(nn.Module):
    def __init__(self, kernel_size: int = 3, **kwargs):
        """Init a basic pooling module.

        - PoolFormer: https://arxiv.org/abs/2111.11418

        Parameters
        ----------
            kernel_size_size : int, default=3
                The pooling kernel_size.
        """
        super().__init__()
        padding = (kernel_size - 1) // 2
        self.pool = nn.AvgPool2d(
            kernel_size, stride=1, padding=padding, count_include_pad=False
        )

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass of the pooling module."""
        y = self.pool(x)

        return y - x


class TokenMixer(nn.Module):
    def __init__(self, name: str, mixer_kwargs: Dict[str, Any], **kwargs) -> None:
        """Token mixer wrapper class.

        Parameters
        ----------
            name : str
                Name of the token-mixer. Allowed: "pool", "self-attention", "mscan",
                "identity", "mlp"
        """
        super().__init__()
        allowed = list(MIXER_LOOKUP.keys())
        if name not in allowed:
            raise ValueError(
                f"Illegal token mixer given. Allowed: {allowed}. Got: '{name}'"
            )

        try:
            self.mixer = MIXER_LOOKUP[name](**mixer_kwargs)
        except Exception as e:
            raise Exception(
                "Encountered an error when trying to init token mixer: "
                f"TokenMixer(name='{name}'): {e.__class__.__name__}: {e}"
            )

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass of a token mixer."""
        return self.mixer(x, **kwargs)


class TokenMixerBlock(nn.Module):
    def __init__(
        self,
        token_mixer: str,
        normalization: str,
        residual: bool = True,
        norm_kwargs: Dict[str, Any] = None,
        mixer_kwargs: Dict[str, Any] = None,
        **kwargs,
    ) -> None:
        """Token mixer block.

        I.e. norm(x) -> tokenmixer(x) -> residual -> (reshape for MLP)

        Parameters
        ----------
            token_mixer : str
                Name of the token mixer. Allowed: "pool", "self-attention", "mscan",
                "identity", "mlp".
            normalization : str
                Name of the normalization method. Allowed: "bn", "bcn", "gn", "in",
                "ln", None.
            residual : bool, default=True
                Flag, whether to use a residual connection at the end of the mixer.
            norm_kwargs : Dict[str, Any], optional
                Arbitrary key-word arguments for the normalization method.
            mixer_kwargs : Dict[str, Any], optional
                Arbitrary key-word arguments for the token mixer module.
        """
        super().__init__()
        self.residual = residual
        norm_kwargs = norm_kwargs if norm_kwargs is not None else {}
        mixer_kwargs = mixer_kwargs if mixer_kwargs is not None else {}
        self.norm = Norm(normalization, **norm_kwargs)
        self.token_mixer = TokenMixer(token_mixer, mixer_kwargs)
        self.reshape = RESHAPE_LOOKUP[token_mixer]  # (B, C, H, W) -> (B, N, C)

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass of the token-mixer block.

        Parameters
        ----------
            x : torch.Tensor
                Input features of shape (B, C, H, W) or (B, N, C).
            **kwargs:
                Arbitrary key-word arguments such e.g. `context` for cross-attn.
        """
        if self.reshape:
            B, C, H, W = x.shape

        if self.residual:
            residual = x

        x = self.norm(x)
        x = self.token_mixer(x, **kwargs)

        if self.residual:
            x = x + residual

        if self.reshape:
            x = x.permute(0, 2, 3, 1).reshape(B, H * W, C)

        return x


MIXER_LOOKUP = {
    "pool": Pooling,
    "mscan": MSCAN,
    "mlp": Mlp,
    "self-attention": SelfAttention,
    "identity": Identity,
}

RESHAPE_LOOKUP = {
    "pool": True,
    "mscan": True,
    "mlp": False,
    "self-attention": False,
    "identity": False,
}



import torch
import torch.nn as nn
import torch.nn.functional as F

from cellseg_models_pytorch.utils import filter2D, sobel_hv


def grad_mse(yhat: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """Compute the gradient mse loss.

    Parameters
    ----------
        yhat : torch.Tensor
            Input tensor of size (B, 2, H, W). Regressed HoVer map.
        target : torch.Tensor
            Target tensor of shape (B, 2, H, W). GT HoVer-map.

    Returns
    -------
        torch.Tensor:
            Computed gradient mse loss matrix. Shape (B, H, W).

    """
    kernel = sobel_hv(window_size=5, device=yhat.device)
    grad_yhat = filter2D(yhat, kernel)
    grad_target = filter2D(target, kernel)
    msge = F.mse_loss(grad_yhat, grad_target, reduction="none")

    return msge


class GradMSE(nn.Module):
    def __init__(self, **kwargs) -> None:
        """Gradient MSE loss.

        Used in the horizontal and vertical gradient branch of HoVer-Net.
        https://arxiv.org/abs/1812.06499
        """
        super().__init__()
        self.eps = 1e-6

    def forward(
        self,
        yhat: torch.Tensor,
        target: torch.Tensor,
        target_inst: torch.Tensor = None,
        **kwargs
    ) -> torch.Tensor:
        """Compute the msge-loss.

        Parameters
        ----------
            yhat : torch.Tensor
                Input tensor (B, 2, H, W). Regressed HoVer map.
            target : torch.Tensor
                Target tensor (B, 2, H, W). GT HoVer map.
            target_inst : torch.Tensor
                instance map target that is used to focus loss to the
                correct objects. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed gradient mse loss (scalar).
        """
        if target_inst is not None:
            focus = torch.stack([target_inst, target_inst], dim=1)
        else:
            focus = torch.ones_like(target)

        loss = focus * grad_mse(yhat, target)
        loss = loss.sum() / focus.clamp_min(self.eps).sum()

        return loss



from pathlib import Path
from typing import Union

import torch
from torch.utils.data import Dataset

from cellseg_models_pytorch.utils import FileHandler

try:
    import tables as tb
except Exception:
    raise ImportError(
        "`pytables` needed for this class. Install with: `pip install tables`"
    )


__all__ = ["HDF5DatasetInfer"]


class HDF5DatasetInfer(Dataset, FileHandler):
    def __init__(self, path: Union[str, Path], n_images: int = None, **kwargs) -> None:
        """Folder dataset that can be used during inference for loading images.

        NOTE: loads only images.

        Parameters
        ----------
            path : str | Path
                Path to the folder containing image files.
            n_images : int, optional
                First n-number of images used from the folder.

        Raises
        ------
            ValueError if the input path has incorrect suffix.
        """
        super().__init__()

        self.path = Path(path)

        if self.path.suffix not in (".h5", ".hdf5"):
            raise ValueError(
                f"The input path has to be a hdf5 db. Got suffix: {self.path.suffix} "
                "Allowed suffices: {('.h5', '.hdf5')}"
            )

        with tb.open_file(self.path) as h5:
            if n_images is not None:
                self.fnames = h5.root.fnames[:n_images]
            else:
                self.fnames = h5.root.fnames[:]

    def __len__(self) -> int:
        """Return the number of items in the db."""
        return len(self.fnames)

    def __getitem__(self, ix: int) -> torch.Tensor:
        """Read image."""
        fn = self.fnames[ix]

        with tb.open_file(self.path.as_posix(), "r") as h5:
            im = h5.root.imgs[ix, ...]

        im = torch.from_numpy(im.transpose(2, 0, 1))
        return {"im": im, "file": Path(fn.decode("UTF-8")).name}



from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn

from .decoder_stage import DecoderStage

__all__ = ["Decoder"]


class Decoder(nn.ModuleDict):
    def __init__(
        self,
        enc_channels: Tuple[int, ...],
        out_channels: Tuple[int, ...] = (256, 128, 64, 32, 16),
        long_skip: Union[None, str, Tuple[str, ...]] = "unet",
        n_conv_layers: Union[None, int, Tuple[int, ...]] = 1,
        n_transformers: Union[None, int, Tuple[int, ...]] = None,
        n_conv_blocks: Union[int, Tuple[Tuple[int, ...], ...]] = 2,
        n_transformer_blocks: Union[int, Tuple[Tuple[int], ...]] = 1,
        stage_params: Optional[Tuple[Dict, ...]] = None,
        style_channels: int = None,
        **kwargs,
    ) -> None:
        """Build a generic U-net-like decoder.

        I.e stack decoder stages that are composed followingly:

        DecoderStage:
            - UpSample(up_method)
            - LongSkip(long_skip_method)
            - ConvLayer (optional)
                - ConvBlock(conv_block_method)
            - TransformerLayer (optional)
                - TransformerBlock(transformer_block_method)

        Parameters
        ----------
            enc_channels : Tuple[int, ...]
                Number of channels at each encoder layer.
            out_channels : Tuple[int, ...], default=(256, 128, 64, 32, 16)
                Number of channels at each decoder layer output.
            long_skip : Union[None, str, Tuple[str, ...]], default="unet"
                long skip method to be used. The argument can be given as a tuple, where
                each value indicates the long-skip method for each stage of the decoder,
                allowing the mixing of long-skip methods in the decoder.
                Allowed: "cross-attn", "unet", "unetpp", "unet3p", "unet3p-lite", None
            n_conv_layers : Union[None, int, Tuple[int, ...]], default=1
                The number of convolution layers inside each of the decoder stages. The
                argument can be given as a tuple, where each value indicates the number
                of conv-layers inside each stage of the decoder allowing the mixing of
                different sized layers inside the stages. If set to None, no conv-layers
                will be included in the decoder.
            n_transformers : Union[None, int, Tuple[int, ...]] , optional
                The number of transformer layers inside each of the decoder stages. The
                argument can be given as a tuple, where each value indicates the number
                of transformer-layers inside each stage of the decoder stages allowing
                the mixing of different sized layers inside the stages. If set to None,
                no transformer layers will be included in the decoder.
            n_conv_blocks : Union[int, Tuple[Tuple[int, ...], ...]], default=2
                The number of blocks inside each conv-layer at each decoder stage. The
                argument can be given as a nested tuple, where each value indicates the
                number of `ConvBlock`s inside a single `ConvLayer` allowing different
                sized blocks inside each conv-layer in the decoder.
            n_transformer_blocks : Union[int, Tuple[Tuple[int], ...]], default=1
                The number of transformer blocks inside each transformer-layer at each
                decoder stage. The argument can be given as a nested tuple, where each
                value indicates the number of `SelfAttention`s inside a single
                `TranformerLayer` allowing different sized transformer blocks inside
                each transformer-layer in the decoder.
            stage_params : Optional[Tuple[Dict, ...]], default=None
                The keyword args for each of the distinct decoder stages. Incudes the
                parameters for the long skip connections, convolutional layers of the
                decoder and transformer layers itself. See the `DecoderStage`
                documentation for more info.
            style_channels : int, default=None
                Number of style vector channels. If None, style vectors are ignored.
                If `n_conv_layers` is None, this is ignored since style vectors are
                applied inside `ConvBlocks`.

        Raises
        ------
            ValueError:
                If there is a mismatch between encoder and decoder channel lengths.
        """
        super().__init__()

        if not len(out_channels) == len(enc_channels):
            raise ValueError(
                "The number of encoder channels need to match the number of "
                f"decoder channels. Got {len(out_channels)} decoder channels "
                f"and {len(enc_channels)} encoder channels."
            )

        out_channels = [enc_channels[0]] + list(out_channels)
        skip_channels = enc_channels[1:]

        # scaling factor assumed to be 2 for the spatial dims and the input
        # has to be divisible by 32. 256 used here just for convenience.
        self.depth = len(out_channels)
        out_dims = [256 // 2**i for i in range(self.depth)][::-1]

        # set layer-level tuple-args
        self.long_skips = self._layer_tuple(long_skip)
        n_conv_layers = self._layer_tuple(n_conv_layers)
        n_transformers = self._layer_tuple(n_transformers)

        # set block-level tuple-args
        n_conv_blocks = self._block_tuple(n_conv_blocks, n_conv_layers)
        n_transformer_blocks = self._block_tuple(n_transformer_blocks, n_transformers)

        # Build decoder
        for i in range(self.depth - 1):
            decoder_block = DecoderStage(
                stage_ix=i,
                dec_channels=tuple(out_channels),
                dec_dims=tuple(out_dims),
                skip_channels=skip_channels,
                long_skip=self._tup_arg(self.long_skips, i),
                n_conv_layers=self._tup_arg(n_conv_layers, i),
                n_conv_blocks=self._tup_arg(n_conv_blocks, i),
                n_transformers=self._tup_arg(n_transformers, i),
                n_transformer_blocks=self._tup_arg(n_transformer_blocks, i),
                style_channels=style_channels,
                **stage_params[i] if stage_params is not None else {"k": None},
            )
            self.add_module(f"decoder_stage{i + 1}", decoder_block)

        self.out_channels = decoder_block.out_channels

    def _tup_arg(self, tup: Tuple[Any, ...], ix: int) -> Union[None, int, str]:
        """Return None if given tuple-arg is None, else, return the value at ix."""
        ret = None
        if tup is not None:
            ret = tup[ix]
        return ret

    def _layer_tuple(
        self, arg: Union[None, str, int, Tuple[Any, ...]]
    ) -> Union[None, Tuple[Any, ...]]:
        """Return a non-nested tuple or None for layer-related arguments."""
        ret = None
        if isinstance(arg, (list, tuple)):
            ret = tuple(arg)
        elif isinstance(arg, (str, int)):
            ret = tuple([arg] * self.depth)
        elif arg is None:
            ret = ret
        else:
            raise ValueError(
                f"Given arg: {arg} should be None, str, int or a Tuple of ints or strs."
            )

        return ret

    def _block_tuple(
        self,
        arg: Union[int, None, Tuple[Tuple[int, ...], ...]],
        n_layers: Tuple[int, ...],
    ) -> Union[None, Tuple[Tuple[int, ...], ...]]:
        """Return a nested tuple or None for block-related arguments."""
        ret = None
        if isinstance(arg, (list, tuple)):
            if not all([isinstance(a, (tuple, list)) for a in arg]):
                raise ValueError(
                    f"Given arg: {arg} should be a nested sequence. Got: {arg}."
                )
            ret = tuple(arg)
        elif isinstance(arg, int):
            if n_layers is not None:
                ret = tuple([tuple([arg] * i) for i in n_layers])
            else:
                ret = None
        elif arg is None:
            ret = ret
        else:
            raise ValueError(f"Given arg: {arg} should be None, int or a nested tuple.")

        return ret

    def forward_features(
        self, features: Tuple[torch.Tensor], style: torch.Tensor = None
    ) -> List[torch.Tensor]:
        """Forward pass of the decoder. Returns all the decoder stage feats."""
        head = features[0]
        skips = features[1:]
        extra_skips = [head] if self.long_skips[0] == "unet3p" else []
        ret_feats = []

        x = head
        for i, decoder_stage in enumerate(self.values()):
            x, extra = decoder_stage(
                x, skips=skips, extra_skips=extra_skips, style=style
            )

            if self.long_skips[i] == "unetpp":
                extra_skips = extra
            elif self.long_skips[i] == "unet3p":
                extra_skips.append(x)

            ret_feats.append(x)

        return ret_feats

    def forward(
        self, *features: Tuple[torch.Tensor], style: torch.Tensor = None
    ) -> torch.Tensor:
        """Forward pass of the decoder."""
        dec_feats = self.forward_features(features, style)

        return dec_feats



from typing import Any, Dict

import torch
import torch.nn as nn

from .base_modules import Norm

__all__ = ["ContiguousEmbed", "PatchEmbed"]


class ContiguousEmbed(nn.Module):
    def __init__(
        self,
        in_channels: int,
        patch_size: int = 1,
        stride: int = 1,
        kernel_size: int = None,
        pad: int = 0,
        head_dim: int = 64,
        num_heads: int = 8,
        flatten: bool = True,
        normalization: str = None,
        norm_kwargs: Dict[str, Any] = None,
        **kwargs,
    ) -> None:
        """Patch an image with nn.Conv2d and then embed.

        NOTE:
        The input is patched via nn.Conv2d i.e the patch dimensions are defined by the
        convolution parameters. The default values are set such that every pixel value
        is a patch. For big inputs this results in OOM errors when computing attention.

        If there is a need for bigger patches with no overlap, you can set for example
        `patch_size = 16` and `stride = 16` to get patches of size 16**2.

        - Input shape: (B, C, H, W)
        - Output shape: (B, H'*W', head_dim*num_heads)
            (If `patch_size=1` & `stride=1` -> H'*W'=H*W).

        NOTE: Optional normalization of the input before patching and projecting.

        Parameters
        ----------
            in_channels : int
                Number of input channels in the input tensor. (3 for RGB).
            patch_size : int, default=1
                Size of the patch. Defaults to 1, meaning that every pixel is a patch.
                (Given that stride is equal to 1.) If `kernel_size` is given, this will
                be ignored.
            stride : int, default=1
                The sliding window stride. Defaults to 1, meaning that every pixel is a
                patch. (Given that patch_size is equal to 1).
            kernel_size : int, optional
                The kernel size for the convolution. If None, the `patch_size` is used.
            pad : int, default=0
                Size of the padding.
            head_dim : int, default=64
                Number of channels per each head.
            num_heads : int, default=8
                Number of heads in multi-head self-attention.
            flatten : bool, default=True
                If True, the output will be flattened to a sequence. After flattening
                output will have shape (B, H'*W', head_dim*num_heads). If False, the
                output shape will remain (B, C, H', W').
            normalization : str, optional
                The name of the normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", "lrn", None
            **norm_kwargs : Dict[str, Any]
                key-word args for the normalization layer. Ignored if normalization
                is None.

        Examples
        --------
            >>> x = torch.rand([1, 3, 256, 256])

            >>> # per-pixel patches of shape 256*256
            >>> conv_patch = ContiguousEmbed(
                    in_channels=3,
                    patch_size=1,
                    stride=1,
                )
            >>> print(conv_patch(x).shape)
            >>> # torch.Size([1, 65536, 512])

            >>> # 16*16 patches
            >>> conv_patch2 = ContiguousEmbed(
                    in_channels=3,
                    patch_size=16,
                    stride=16,
                )
            >>> print(conv_patch2(x).shape)
            >>> # torch.Size([1, 256, 512])

            >>> # Downsampling input to patches of shape 64*64
            >>> conv_patch3 = ContiguousEmbed(
                    in_channels=3,
                    stride=4,
                    kernel_size=7,
                    pad=2
                )
            >>> print(conv_patch3(x).shape)
            >>> # torch.Size([1, 4096, 512])
        """
        super().__init__()
        self.flatten = flatten
        self.proj_dim = head_dim * num_heads
        self.kernel_size = patch_size if kernel_size is None else kernel_size
        self.pad = pad
        self.stride = stride
        norm_kwargs = norm_kwargs if norm_kwargs is not None else {}

        self.norm = Norm(normalization, **norm_kwargs)
        self.proj = nn.Conv2d(
            in_channels,
            self.proj_dim,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.pad,
        )

    def get_patch_size(self, img_size: int) -> int:
        """Get the patch size from the conv params."""
        return int(
            (((img_size + 2 * self.pad - (self.kernel_size - 1)) - 1) / self.stride) + 1
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for projection."""
        B, _, H, W = x.shape

        # 1. Normalize
        x = self.norm(x)

        # 2. Patch and project.
        x = self.proj(x)  # (B, proj_dim, H', W')

        # 3. reshape to a sequence.
        # Every patch has been projected into a `proj_dim` long vector.
        if self.flatten:
            p_H = self.get_patch_size(H)
            p_W = self.get_patch_size(W)

            # flatten
            x = x.permute(0, 2, 3, 1).reshape(
                B, p_H * p_W, self.proj_dim
            )  # (B, H'*W', proj_dim)

        return x


class PatchEmbed(nn.Module):
    def __init__(
        self,
        in_channels: int,
        patch_size: int = 16,
        head_dim: int = 64,
        num_heads: int = 8,
        normalization: int = None,
        **norm_kwargs,
    ) -> None:
        """Patch an input image and then embed/project.

        NOTE: This implementation first patches the input image by reshaping it
        and then embeds/projects it with nn.Linear.

        NOTE: Optional normalization of the input before patching and projecting.

        - Input shape: (B, C, H, W)
        - Patched shape: (B, H//patch_size * W//patch_size, C*patch_size**2)
        - Embedded output shape: (B, H//patch_size * W//patch_size, head_dim*num_heads)

        Parameters
        ----------
            in_channels : int
                Number of input channels in the input tensor.
            patch_size : int, default=16
                The H and W size of the patch.
            head_dim : int, default=64
                Number of channels per each head.
            num_heads : int, default=8
                Number of heads in multi-head self-attention.
            normalization : str, optional
                The name of the normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", "lrn", None
            **norm_kwargs : Dict[str, Any]
                key-word args for the normalization layer. Ignored if normalization
                is None.

        Examples
        --------
            >>> x = torch.rand([1, 3, 256, 256])

            >>> # patches of shape 16*16
            >>> lin_patch = PatchEmbed(
                    in_channels=3,
                    patch_size=16,
                )
            >>> print(lin_patch(x).shape)
            >>> # torch.Size([1, 256, 512])

        """
        super().__init__()
        self.proj_dim = head_dim * num_heads
        self.patch_size = patch_size
        self.norm = Norm(normalization, **norm_kwargs)
        self.proj = nn.Linear(in_channels * (patch_size**2), self.proj_dim)

    def img2patch(self, x: torch.Tensor) -> torch.Tensor:
        """Patch an input image of shape (B, C, H, W).

        Adapted from: PyTorch Lightning ViT tutorial.

        Parameters
        ----------
            x : torch.Tensor
                Input image of shape (B, C, H, W).

        Returns
        -------
            torch.Tensor:
                Patched and flattened input image.
                Shape: (B, H//patch_size * W//patch_size, C*patch_size**2)
        """
        B, C, H, W = x.shape
        x = x.reshape(
            B,
            C,
            H // self.patch_size,
            self.patch_size,
            W // self.patch_size,
            self.patch_size,
        )  # (B, C, H', patch_size, W', patch_size)

        x = x.permute(0, 2, 4, 1, 3, 5)  # (B, H', W', C, p_H, p_W)
        x = x.flatten(1, 2)  # (B, H'*W', C, p_H, p_W)
        x = x.flatten(2, 4)  # (B, H'*W', C*p_H*p_W)

        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward patch embedding."""
        # 1. Normalize
        x = self.norm(x)  # (B, C, H, W)

        # 2. Patch
        x = self.img2patch(x)  # (B, H//patch_size * W//patch_size, C*patch_size**2)

        # 3. Project/Embed
        x = self.proj(x)

        return x



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
from tqdm import tqdm

from ...utils import FileHandler
from ._base_writer import BaseWriter

__all__ = ["FolderWriter"]


class FolderWriter(BaseWriter):
    def __init__(
        self,
        in_dir_im: str,
        save_dir_im: str,
        in_dir_mask: Optional[str] = None,
        save_dir_mask: Optional[str] = None,
        patch_size: Optional[Tuple[int, int]] = None,
        stride: Optional[int] = None,
        transforms: Optional[List[str]] = None,
    ) -> None:
        """Write overlapping patches to a folder from image and .mat files.

        Image patches will be written as .png and mask patches as .mat files.

        NOTE: Very ad-hoc and not tested so there exists a chance of failure...

        Parameters
        ----------
            in_dir_im : str
                Path to the folder of images
            save_dir_im : str
                Path to the folder where the new imgs will be saved.
            in_dir_mask : str, optional
                Path to the folder of masks.
            save_dir_mask : str, optional
                Path to the folder where the new masks will be saved.
            patch_size : Tuple[int, int], optional
                Height and width of the extracted patches.
            stride : int, optional
                Stride for the sliding window.
            transforms : List[str], optional
                A list of transforms to apply to the images and masks. Allowed ones:
                "blur", "non_spatial", "non_rigid", "rigid", "hue_sat", "random_crop",
                "center_crop", "resize"

        Raises
        ------
            ValueError if issues with given paths or the files in those folders.

        Example
        -------
            >>> # Patch and write 2 folder
            >>> writer = FolderWriter(
                    "/path/to/my/imgs/,
                    "/path/to/my/masks/,
                    "/save/img_patches/here/,
                    "/save/mask_patches/here/,
                    patch_size=(320, 320),
                    stride=160,
                    transforms=["rigid"]
                )
            >>> writer.write(tiling=True)

            >>> # Don't patch, just write 2 folder
            >>> writer = FolderWriter(
                    "/path/to/my/imgs/,
                    "/path/to/my/masks/,
                    "/save/img_patches/here/,
                    "/save/mask_patches/here/,
                    transforms=["rigid"]
                )
            >>> writer.write(tiling=False)

        """
        super().__init__(
            in_dir_im=in_dir_im,
            in_dir_mask=in_dir_mask,
            stride=stride,
            patch_size=patch_size,
            transforms=transforms,
        )
        self.save_dir_im = Path(save_dir_im)

        if save_dir_mask is not None:
            self.save_dir_mask = Path(save_dir_mask)

    def write(
        self, tiling: bool = False, pre_proc: Callable = None, msg: str = None
    ) -> None:
        """Write patches to a folder.

        Parameters
        ----------
            tiling : bool, default=False
                Apply tiling to the images before saving.
            pre_proc : Callable, optional
                An optional pre-processing function for the masks.
        """
        it = self.fnames_im
        if self.fnames_mask is not None:
            it = zip(self.fnames_im, self.fnames_mask)

        with tqdm(it, total=len(self.fnames_im)) as pbar:
            total_tiles = 0
            for fn in pbar:
                # get the img and masks filenames
                if self.fnames_mask is not None:
                    fn_im, fn_mask = fn
                else:
                    fn_im = fn
                    fn_mask = None

                msg = msg if msg is not None else ""
                pbar.set_description(f"Extracting {msg} patches to folders..")

                # optionally patch and process images and masks
                im, masks = self.get_array(
                    fn_im, fn_mask, tiling=tiling, pre_proc=pre_proc
                )

                n_tiles = im.shape[0] if tiling else 1
                arg_list = []
                for i in range(n_tiles):
                    kw = self._set_kwargs(im, fn_im, masks, fn_mask, tiling, i)
                    arg_list.append(kw)

                self._write_parallel(self._save2folder, arg_list)
                total_tiles += n_tiles
                pbar.set_postfix_str(f"# of extracted tiles {total_tiles}")

    def _set_kwargs(
        self,
        im: np.ndarray,
        img_path: Path,
        masks: Dict[str, np.ndarray] = None,
        mask_path: Path = None,
        tiling: bool = False,
        ix: int = None,
    ) -> Dict[str, Any]:
        """Define one set of arguments for saving function."""
        fn = img_path.with_suffix("").name

        fn_im_new = f"{fn}.png"
        fn_mask_new = None
        if mask_path is not None:
            fn_mask_new = f"{fn}.mat"

        if tiling:
            fn_im_new = f"{fn}_patch{ix + 1}.png"

            if mask_path is not None:
                fn_mask_new = f"{fn}_patch{ix + 1}.mat"

        im_save_path = self.save_dir_im / fn_im_new

        mask_save_path = None
        if mask_path is not None:
            mask_save_path = self.save_dir_mask / fn_mask_new

        kwarg = {"path_im": im_save_path, "path_mask": mask_save_path}

        if tiling:
            kwarg["image"] = im[ix : ix + 1].squeeze()

            if mask_path is not None:
                for k, m in masks.items():
                    kwarg[k] = m[ix : ix + 1].squeeze()
        else:
            kwarg["image"] = im
            if mask_path is not None:
                for k, m in masks.items():
                    kwarg[k] = m

        return kwarg

    def save2folder(
        self,
        path_im: str,
        image: np.ndarray,
        path_mask: str = None,
        inst_map: np.ndarray = None,
        type_map: np.ndarray = None,
        sem_map: np.ndarray = None,
    ) -> None:
        """Write image and corresponding masks to folder."""
        FileHandler.write_img(path_im, image)
        if path_mask is not None:
            FileHandler.write_mat(path_mask, inst_map, type_map, sem_map)

    def _save2folder(self, kwargs) -> None:
        """Unpack the kwargs for `save2folder`."""
        return self.save2folder(**kwargs)



from .base._multitask_unet import MultiTaskUnet
from .cellpose.cellpose import (
    CellPoseUnet,
    cellpose_base,
    cellpose_plus,
    omnipose_base,
    omnipose_plus,
)
from .hovernet.hovernet import (
    HoverNet,
    hovernet_base,
    hovernet_plus,
    hovernet_small,
    hovernet_small_plus,
)
from .stardist.stardist import (
    StarDistUnet,
    stardist_base,
    stardist_base_multiclass,
    stardist_plus,
)

MODEL_LOOKUP = {
    "cellpose_base": cellpose_base,
    "cellpose_plus": cellpose_plus,
    "omnipose_base": omnipose_base,
    "omnipose_plus": omnipose_plus,
    "hovernet_base": hovernet_base,
    "hovernet_small": hovernet_small,
    "hovernet_small_plus": hovernet_small_plus,
    "stardist_base": stardist_base,
    "stardist_plus": stardist_plus,
    "stardist_base_multiclass": stardist_base_multiclass,
}


def get_model(name: str, type: str, ntypes: int = None, ntissues: int = None):
    """Get the corect model at hand given name and type."""
    if name == "stardist":
        if type == "base":
            model = MODEL_LOOKUP["stardist_base_multiclass"](
                n_rays=32, type_classes=ntypes
            )
        elif type == "plus":
            model = MODEL_LOOKUP["stardist_plus"](
                n_rays=32, type_classes=ntypes, sem_classes=ntissues
            )
    elif name == "cellpose":
        if type == "base":
            model = MODEL_LOOKUP["cellpose_base"](type_classes=ntypes)
        elif type == "plus":
            model = MODEL_LOOKUP["cellpose_plus"](
                type_classes=ntypes, sem_classes=ntissues
            )
    elif name == "omnipose":
        if type == "base":
            model = MODEL_LOOKUP["omnipose_base"](type_classes=ntypes)
        elif type == "plus":
            model = MODEL_LOOKUP["omnipose_plus"](
                type_classes=ntypes, sem_classes=ntissues
            )
    elif name == "hovernet":
        if type == "base":
            model = MODEL_LOOKUP["hovernet_base"](type_classes=ntypes)
        elif type == "small":
            model = MODEL_LOOKUP["hovernet_small"](type_classes=ntypes)
        elif type == "plus":
            model = MODEL_LOOKUP["hovernet_plus"](
                type_classes=ntypes, sem_classes=ntissues
            )
        elif type == "small_plus":
            model = MODEL_LOOKUP["hovernet_small_plus"](
                type_classes=ntypes, sem_classes=ntissues
            )
    else:
        raise ValueError("Unknown model type or name.")

    return model


__all__ = [
    "MultiTaskUnet",
    "HoverNet",
    "hovernet_base",
    "hovernet_plus",
    "hovernet_small",
    "hovernet_small_plus",
    "CellPoseUnet",
    "cellpose_base",
    "cellpose_plus",
    "omnipose_base",
    "omnipose_plus",
    "StarDistUnet",
    "stardist_base",
    "stardist_plus",
    "stardist_base_multiclass",
    "MODEL_LOOKUP",
    "get_model",
]



from typing import Tuple

import torch
import torch.nn as nn

from ...modules import Identity
from .merging import Merge

__all__ = ["UnetSkip"]


class UnetSkip(nn.Module):
    def __init__(
        self,
        stage_ix: int,
        in_channels: int,
        skip_channels: Tuple[int, ...] = None,
        merge_policy: str = "sum",
        **kwargs
    ) -> None:
        """U-net-like skip connection block.

        U-Net: Convolutional Networks for Biomedical Image Segmentation
            - https://arxiv.org/abs/1505.04597#

        Parameters
        ----------
            stage_ix : int
                Index number signalling the current decoder stage
            in_channels : int, default=None
                The number of channels in the input tensor.
            skip_channels : Tuple[int, ...]
                Tuple of the number of channels in the encoder stages.
                Order is bottom up. This list does not include the final
                bottleneck stage out channels. e.g. (1024, 512, 256, 64).
            merge_policy : str, default="sum"
                Sum or concatenate the features together. One of ("sum", "cat").
        """
        super().__init__()
        self.merge_policy = merge_policy
        self.in_channels = in_channels
        self.stage_ix = stage_ix

        self.skip_out_chl = 0
        self.merge = Identity()
        if stage_ix < len(skip_channels):
            self.skip_out_chl = skip_channels[stage_ix]
            self.merge = Merge(
                self.merge_policy,
                in_channels=self.in_channels,
                skip_channels=(self.skip_out_chl,),
                **kwargs
            )

    @property
    def out_channels(self) -> int:
        """Out channels."""
        out_channels = self.in_channels
        if self.merge_policy == "cat":
            out_channels += self.skip_out_chl

        return out_channels

    def forward(
        self, x: torch.Tensor, skips: Tuple[torch.Tensor, ...], **kwargs
    ) -> Tuple[torch.Tensor, None]:
        """Forward pass of the U-net long skip connection.

        Parameters
        ----------
            x : torch.Tensor
                Input from the previous decoder layer. Shape (B, C, H, W).
            skips : Tuple[torch.Tensor, ...]
                All the encoder feature maps. Shapes: (B, C, H, W).
            ix : int
                Index of the decoder stage.

        Returns
        -------
            Tuple[torch.Tensor, ...]:
                The skip connection tensor. Shape (B, C ( + n_skip_cahnnels), H, W).
        """
        if self.stage_ix < len(skips):
            skip = skips[self.stage_ix]
            x = self.merge(x, (skip,))

        return x



from typing import Any, Callable, Generator, List, Union

from pathos.pools import ProcessPool, SerialPool, ThreadPool

__all__ = ["run_pool"]


def iter_pool_generator(it: Generator, res: List = None) -> Union[List[Any], None]:
    """Iterate over a pool generator object.

    Parameters
    ----------
        it : Generator
            A Generator object containing results from a concurrent run.
        res : List | None
            An empty list, where the results from the generator will be saved.
            If None, no results will be saved.

    Returns
    -------
        Union[List[Any], None]:
            A list of results or None.
    """
    if res is not None:
        for x in it:
            res.append(x)
    else:
        for _ in it:
            pass

    return res


def run_pool(
    func: Callable,
    args: List[Any],
    ret: bool = True,
    pooltype: str = "thread",
    maptype: str = "amap",
) -> Union[List[Any], None]:
    """Run a pathos Thread, Process or Serial pool object.

    NOTE: if `ret` is set to True and `func` callable does not return anything. This
          will return a list of None values.

    Parameters
    ----------
        func : Callable
            The function that will be copied to existing cores and run in parallel.
        args : List[Any]
            A list of arguments for each of the parallelly executed functions.
        ret : bool, default=True
            Flag, whether to return a list of results from the pool object. Will be set
            to False e.g. when saving data to disk in parallel etc.
        pooltype : str, default="thread"
            The pathos pooltype. Allowed: ("process", "thread", "serial")
        maptype : str, default="amap"
            The map type of the pathos Pool object.
            Allowed: ("map", "amap", "imap", "uimap")

    Raises
    ------
        ValueError: if illegal `pooltype` or `maptype` is given.

    Returns
    -------
        Union[List[Any], None]:
            A list of results or None.

    Example
    -------
        >>> f = myfunc # any function.
        >>> args = (1, 2, 3)
        >>> res_list = run_pool(f, args)
    """
    allowed = ("process", "thread", "serial")
    if pooltype not in allowed:
        raise ValueError(f"Illegal `pooltype`. Got {pooltype}. Allowed: {allowed}")

    allowed = ("map", "amap", "imap", "uimap")
    if maptype not in allowed:
        raise ValueError(f"Illegal `maptype`. Got {maptype}. Allowed: {allowed}")

    Pool = None
    if pooltype == "thread":
        Pool = ThreadPool
    elif pooltype == "process":
        Pool = ProcessPool
    else:
        if maptype in ("amap", "uimap"):
            raise ValueError(
                f"`SerialPool` has only `map` & `imap` implemented. Got: {maptype}."
            )
        Pool = SerialPool

    results = [] if ret else None
    if maptype == "map":
        with Pool() as pool:
            it = pool.map(func, args)
            results = iter_pool_generator(it, results)
    elif maptype == "amap":
        with Pool() as pool:
            results = pool.amap(func, args).get()
    elif maptype == "imap":
        with Pool() as pool:
            it = pool.imap(func, args)
            results = iter_pool_generator(it, results)
    elif maptype == "uimap":
        with Pool() as pool:
            it = pool.uimap(func, args)
            results = iter_pool_generator(it, results)

    return results



from typing import Any, Dict, Optional, Tuple

import torch
import torch.nn as nn

from ...modules import Transformer2D
from ...modules.patch_embeddings import ContiguousEmbed

__all__ = ["CrossAttentionSkip"]


class CrossAttentionSkip(nn.Module):
    def __init__(
        self,
        stage_ix: int,
        in_channels: int,
        skip_channels: Tuple[int, ...] = None,
        num_heads: int = 8,
        head_dim: int = 64,
        n_blocks: int = 1,
        block_types: Tuple[str, ...] = ("exact",),
        computation_types: Tuple[str, ...] = ("basic",),
        dropouts: Tuple[float, ...] = (0.0,),
        biases: Tuple[bool, ...] = (False,),
        layer_scales: Tuple[bool, ...] = (False,),
        activation: str = "star_relu",
        mlp_ratio: int = 2,
        slice_size: int = 4,
        patch_embed_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        """Skip connection (U-Net-like) via cross-attention.

        Does the long skip connection through a cross-attention transformer rather than
        merging or summing the skip features to the upsampled decoder feature-map.

        Parameters
        ----------
            stage_ix : int
                Index number signalling the current decoder stage
            in_channels : int, default=None
                The number of channels in the input tensor.
            skip_channels : Tuple[int, ...]
                Tuple of the number of channels in the encoder stages.
                Order is bottom up. This list does not include the final
                bottleneck stage out channels.
            num_heads : int, default=8
                Number of heads in multi-head attention.
            head_dim : int, default=64
                The out dim of the heads.
            n_blocks : int, default=1
                Number of SelfAttentionBlocks used in this layer.
            block_types : Tuple[str, ...], default=("exact", )
                The name of the SelfAttentionBlocks in the TransformerLayer.
                Length of the tuple has to equal `n_blocks`
                Allowed names: "basic". "slice", "flash".
            computation_types : Tuple[str, ...], default=("basic", )
                The way of computing the attention matrices in the SelfAttentionBlocks
                in the TransformerLayer. Length of the tuple has to equal `n_blocks`
                Allowed styles: "basic". "slice", "flash", "memeff", "slice_memeff".
            dropouts : Tuple[float, ...], default=(False, )
                Dropout probabilities for the SelfAttention blocks.
            biases : bool, default=(True, True)
                Include bias terms in the SelfAttention blocks.
            layer_scales : bool, default=(False, )
                Learnable layer weights for the self-attention matrix.
            activation : str, default="star_relu"
                The activation function applied at the end of the transformer layer fc.
                One of ("geglu", "approximate_gelu", "star_relu").
            mlp_ratio : int, default=4
                Multiplier that defines the out dimension of the final fc projection
                layer.
            slice_size : int, default=4
                Slice size for sliced self-attention. This is used only if
                `name = "slice"` for a SelfAttentionBlock.
            patch_embed_kwargs: Dict[str, Any], optional
                Extra key-word arguments for the patch embedding module. See the
                `ContiguousEmbed` module for more info.
        """
        super().__init__()
        self.in_channels = in_channels
        self.stage_ix = stage_ix

        if stage_ix < len(skip_channels):
            context_channels = skip_channels[stage_ix]

            self.context_patch_embed = ContiguousEmbed(
                in_channels=context_channels,
                patch_size=1,
                num_heads=num_heads,
                head_dim=head_dim,
                normalization="gn",
                norm_kwargs={"num_features": context_channels},
                **patch_embed_kwargs if patch_embed_kwargs is not None else {},
            )

            self.tranformer = Transformer2D(
                in_channels=in_channels,
                cross_attentions_dims=self.context_patch_embed.proj_dim,
                num_heads=num_heads,
                head_dim=head_dim,
                n_blocks=n_blocks,
                block_types=block_types,
                computation_types=computation_types,
                dropouts=dropouts,
                biases=biases,
                layer_scales=layer_scales,
                activation=activation,
                slice_size=slice_size,
                mlp_ratio=mlp_ratio,
                patch_embed_kwargs=patch_embed_kwargs,
                **kwargs,
            )

    @property
    def out_channels(self) -> int:
        """Out channels."""
        return self.in_channels

    def forward(
        self, x: torch.Tensor, skips: Tuple[torch.Tensor], **kwargs
    ) -> torch.Tensor:
        """Forward pass of the skip connection."""
        if self.stage_ix < len(skips):
            context = skips[self.stage_ix]  # (B, C, H, W)

            # embed context for cross-attm transformer: (B, H'*W', num_heads*head_dim)
            context = self.context_patch_embed(context)

            x = self.tranformer(x, context=context)  # (B, C, H, W)

        return x



"""
Ported from hovernet repo.

https://github.com/vqdang/hover_net/blob/master/models/hovernet/targets.py

MIT License

Copyright (c) 2020 vqdang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as np
from scipy import ndimage as ndi

from ...utils import bounding_box, remove_small_objects

__all__ = ["gen_hv_maps"]


def gen_hv_maps(inst_map: np.ndarray, min_size: int = 5) -> np.ndarray:
    """Generate horizontal and vertical gradient maps from label mask.

    Hover-Net: Simultaneous segmentation and classification of nuclei in multi-tissue
    histology images, Simon Graham, Quoc Dang Vu, Shan E Ahmed Raza, Ayesha Azam,
    Yee Wah Tsang, Jin Tae Kwak, Nasir Rajpoot, Medical Image Analysis, Volume 58,
    2019, 101563, ISSN 1361-8415, doi: https://doi.org/10.1016/j.media.2019.101563.

    https://www.sciencedirect.com/science/article/pii/S1361841519301045

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        min_size : int, default=5
            Min size for objects. Objects less than this many pixels
            are removed.

    Returns
    -------
        np.ndarray:
            The Y- and X-gradient maps in this order. Shape (2, H, W).
            Dtype: float64.
    """
    x_map = np.zeros_like(inst_map, dtype=np.float64)
    y_map = np.zeros_like(inst_map, dtype=np.float64)

    # inst_map = remove_small_objects(inst_map, min_size=min_size, out=inst_map)
    inst_map = remove_small_objects(inst_map, min_size=min_size)

    inst_list = list(np.unique(inst_map))
    if 0 in inst_list:
        inst_list.remove(0)

    for inst_id in inst_list:
        inst = np.array(inst_map == inst_id, np.int32)
        y1, y2, x1, x2 = bounding_box(inst)
        y1 = y1 - 2 if y1 - 2 >= 0 else y1
        x1 = x1 - 2 if x1 - 2 >= 0 else x1
        x2 = x2 + 2 if x2 + 2 <= inst_map.shape[1] - 1 else x2
        y2 = y2 + 2 if y2 + 2 <= inst_map.shape[0] - 1 else y2
        inst = inst[y1:y2, x1:x2]

        # instance center of mass, rounded to nearest pixel
        inst_com = list(ndi.center_of_mass(inst))
        inst_com[0] = int(inst_com[0] + 0.5)
        inst_com[1] = int(inst_com[1] + 0.5)

        inst_x_range = np.arange(1, inst.shape[1] + 1)
        inst_y_range = np.arange(1, inst.shape[0] + 1)

        # shifting center of pixels grid to instance center of mass
        inst_x_range -= inst_com[1]
        inst_y_range -= inst_com[0]

        inst_x, inst_y = np.meshgrid(inst_x_range, inst_y_range)

        # remove coord outside of instance
        inst_x[inst == 0] = 0
        inst_y[inst == 0] = 0
        inst_x = inst_x.astype(np.float64)
        inst_y = inst_y.astype(np.float64)

        # normalize min into -1 scale
        if np.min(inst_x) < 0:
            inst_x[inst_x < 0] /= -np.amin(inst_x[inst_x < 0])
        if np.min(inst_y) < 0:
            inst_y[inst_y < 0] /= -np.amin(inst_y[inst_y < 0])
        # normalize max into +1 scale
        if np.max(inst_x) > 0:
            inst_x[inst_x > 0] /= np.amax(inst_x[inst_x > 0])
        if np.max(inst_y) > 0:
            inst_y[inst_y > 0] /= np.amax(inst_y[inst_y > 0])

        ####
        x_map_box = x_map[y1:y2, x1:x2]
        x_map_box[inst > 0] = inst_x[inst > 0]

        y_map_box = y_map[y1:y2, x1:x2]
        y_map_box[inst > 0] = inst_y[inst > 0]

    hover_map = np.stack([y_map, x_map])

    return hover_map



import pytest

from cellseg_models_pytorch.utils.seg_benchmark import SegBenchmarker

classes = {
    "0": 0,
    "1": 1,
    "2": 2,
    "3": 3,
    "4": 4,
    "5": 5,
    "6": 6,
}


@pytest.mark.parametrize("how", ["binary", "multi", None])
@pytest.mark.parametrize("inputf", ["h5", "folder"])
def test_sem_seg_bm(mask_patch_dir, hdf5db, how, inputf):
    if inputf == "folder":
        inpath = mask_patch_dir
    else:
        inpath = hdf5db

    bm = SegBenchmarker(
        true_path=inpath,
        pred_path=inpath,
        type_classes=classes,
        sem_classes=classes,
    )

    if how == "binary":
        res = bm.run_inst_benchmark(how, metrics=("dice2",))
    elif how == "multi":
        res = bm.run_inst_benchmark(how, metrics=("dice2",))
    else:
        res = bm.run_sem_benchmark(metrics=("iou",))

    assert isinstance(res, list)  # convenience



import torch
import torch.nn.functional as F

__all__ = ["filter2D", "gaussian", "gaussian_kernel2d", "sobel_hv"]


def filter2D(input_tensor: torch.Tensor, kernel: torch.Tensor) -> torch.Tensor:
    """Convolves a given kernel on input tensor without losing dimensional shape.

    Parameters
    ----------
        input_tensor : torch.Tensor
            Input image/tensor.
        kernel : torch.Tensor
            Convolution kernel/window.

    Returns
    -------
        torch.Tensor:
            The convolved tensor of same shape as the input.
    """
    (_, channel, _, _) = input_tensor.size()

    # "SAME" padding to avoid losing height and width
    pad = [
        kernel.size(2) // 2,
        kernel.size(2) // 2,
        kernel.size(3) // 2,
        kernel.size(3) // 2,
    ]
    pad_tensor = F.pad(input_tensor, pad, "replicate")

    out = F.conv2d(pad_tensor, kernel, groups=channel)
    return out


def gaussian(
    window_size: int, sigma: float, device: torch.device = None
) -> torch.Tensor:
    """Create a gaussian 1D tensor.

    Parameters
    ----------
        window_size : int
            Number of elements for the output tensor.
        sigma : float
            Std of the gaussian distribution.
        device : torch.device
            Device for the tensor.

    Returns
    -------
        torch.Tensor:
            A gaussian 1D tensor. Shape: (window_size, ).
    """
    x = torch.arange(window_size, device=device).float() - window_size // 2
    if window_size % 2 == 0:
        x = x + 0.5

    gauss = torch.exp((-x.pow(2.0) / float(2 * sigma**2)))

    return gauss / gauss.sum()


def gaussian_kernel2d(
    window_size: int, sigma: float, n_channels: int = 1, device: torch.device = None
) -> torch.Tensor:
    """Create 2D window_size**2 sized kernel a gaussial kernel.

    Parameters
    ----------
        window_size : int
            Number of rows and columns for the output tensor.
        sigma : float
            Std of the gaussian distribution.
        n_channel : int
            Number of channels in the image that will be convolved with
            this kernel.
        device : torch.device
            Device for the kernel.

    Returns:
    -----------
        torch.Tensor:
            A tensor of shape (1, 1, window_size, window_size)
    """
    kernel_x = gaussian(window_size, sigma, device=device)
    kernel_y = gaussian(window_size, sigma, device=device)

    kernel_2d = torch.matmul(kernel_x.unsqueeze(-1), kernel_y.unsqueeze(-1).t())
    kernel_2d = kernel_2d.expand(n_channels, 1, window_size, window_size)

    return kernel_2d


def sobel_hv(window_size: int = 5, device: torch.device = None):
    """Create a kernel that is used to compute 1st order derivatives.

    Parameters
    ----------
        window_size : int
            Size of the convolution kernel.
        device : torch.device:
            Device for the kernel.

    Returns
    -------
        torch.Tensor:
            the computed 1st order derivatives of the input tensor.
            Shape (B, 2, H, W)

    Raises
    ------
        ValueError:
            If `window_size` is not an odd number.
    """
    if not window_size % 2 == 1:
        raise ValueError(f"window_size must be odd. Got: {window_size}")

    # Generate the sobel kernels
    range_h = torch.arange(
        -window_size // 2 + 1, window_size // 2 + 1, dtype=torch.float32, device=device
    )
    range_v = torch.arange(
        -window_size // 2 + 1, window_size // 2 + 1, dtype=torch.float32, device=device
    )
    h, v = torch.meshgrid(range_h, range_v)

    kernel_h = h / (h * h + v * v + 1e-6)
    kernel_h = kernel_h.unsqueeze(0).unsqueeze(0)

    kernel_v = v / (h * h + v * v + 1e-6)
    kernel_v = kernel_v.unsqueeze(0).unsqueeze(0)

    return torch.cat([kernel_h, kernel_v], dim=0)



import cv2
import numpy as np
import scipy.ndimage as ndi
import skimage.morphology as morph
import skimage.segmentation as segm

from cellseg_models_pytorch.utils import (
    binarize,
    naive_thresh_prob,
    percentile_normalize99,
    remove_small_objects,
)


def post_proc_dran(
    inst_map: np.ndarray, contour_map: np.ndarray, **kwargs
) -> np.ndarray:
    """DRAN post-processing pipeline.

    https://www.frontiersin.org/articles/10.3389/fbioe.2019.00053/full

    This is not the original implementation but follows along the steps
    introduced in the paper. Added dilation to the end of the pipeline.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled or binary mask. Shape (H, W).
        contour_map : np.ndarray
            Contour map. Shape (H, W).

    Returns
    -------
        np.ndarray:
            post-processed inst map. Shape (H, W)
    """
    contour = percentile_normalize99(contour_map, amin=0, amax=1)
    cnt_binary = binarize(naive_thresh_prob(contour))
    cnt_binary = cv2.dilate(cnt_binary, morph.disk(3), iterations=1)

    binary = binarize(inst_map)
    binary = remove_small_objects(binary.astype(bool), min_size=10).astype("uint8")

    markers = binary - cnt_binary
    markers[markers != 1] = 0
    markers = ndi.label(markers)[0]

    distance = ndi.distance_transform_edt(markers)
    distance = 255 * (distance / (np.amax(distance) + 1e-7))

    inst_map = segm.watershed(-distance, markers=markers, mask=binary)

    return inst_map



import math
from typing import List, Sequence, Tuple

import numpy as np
from numba import njit, prange
from scipy.spatial import KDTree

from ....utils import intersection

__all__ = ["get_bboxes", "nms_stardist"]


@njit(parallel=True)
def get_bboxes(
    dist: np.ndarray, points: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, int]:
    """Get bounding boxes from the non-zero pixels of the radial distance maps.

    This is basically a translation from the stardist repo cpp code to python

    NOTE: jit compiled and parallelized with numba.

    Parameters
    ----------
        dist : np.ndarray
            The non-zero values of the radial distance maps. Shape: (n_nonzero, n_rays).
        points : np.ndarray
            The yx-coordinates of the non-zero points. Shape (n_nonzero, 2).

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, int]:
        Returns the x0, y0, x1, y1 bbox coordinates, bbox areas and the maximum
        radial distance in the image.
    """
    n_polys = dist.shape[0]
    n_rays = dist.shape[1]

    bbox_x1 = np.zeros(n_polys)
    bbox_x2 = np.zeros(n_polys)
    bbox_y1 = np.zeros(n_polys)
    bbox_y2 = np.zeros(n_polys)

    areas = np.zeros(n_polys)
    angle_pi = 2 * math.pi / n_rays
    max_dist = 0

    for i in prange(n_polys):
        max_radius_outer = 0
        py = points[i, 0]
        px = points[i, 1]

        for k in range(n_rays):
            d = dist[i, k]
            y = py + d * np.sin(angle_pi * k)
            x = px + d * np.cos(angle_pi * k)

            if k == 0:
                bbox_x1[i] = x
                bbox_x2[i] = x
                bbox_y1[i] = y
                bbox_y2[i] = y
            else:
                bbox_x1[i] = min(x, bbox_x1[i])
                bbox_x2[i] = max(x, bbox_x2[i])
                bbox_y1[i] = min(y, bbox_y1[i])
                bbox_y2[i] = max(y, bbox_y2[i])

            max_radius_outer = max(d, max_radius_outer)

        areas[i] = (bbox_x2[i] - bbox_x1[i]) * (bbox_y2[i] - bbox_y1[i])
        max_dist = max(max_dist, max_radius_outer)

    return bbox_x1, bbox_y1, bbox_x2, bbox_y2, areas, max_dist


@njit
def _suppress_bbox(
    query: Sequence[int],
    current_idx: int,
    boxes: np.ndarray,
    areas: np.ndarray,
    suppressed: List[bool],
    iou_threshold: float = 0.5,
) -> np.ndarray:
    """Inner loop of the stardist nms algorithm where bboxes are suppressed.

    NOTE: Numba compiled only for performance.
          Parallelization had only a negative effect on run-time on.
          12-core hyperthreaded Intel(R) Core(TM) i7-9750H CPU @ 2.60GHz.
    """
    for i in range(len(query)):
        query_idx = query[i]

        if suppressed[query_idx]:
            continue

        overlap = intersection(boxes[current_idx], boxes[query_idx])
        iou = overlap / min(areas[current_idx] + 1e-10, areas[query_idx] + 1e-10)
        suppressed[query_idx] = iou > iou_threshold

    return suppressed


def nms_stardist(
    boxes: np.ndarray,
    points: np.ndarray,
    scores: np.ndarray,
    areas: np.ndarray,
    max_dist: float,
    score_threshold: float = 0.5,
    iou_threshold: float = 0.5,
) -> np.ndarray:
    """Non maximum suppression for stardist bboxes.

    NOTE: This implementation relies on `scipy.spatial` `KDTree`

    NOTE: This version of nms is faster than the original one in stardist repo
    and is fully written in python. The differenecs in the resulting instance
    segmentation masks are neglible.

    Parameters
    ----------
        boxes : np.ndarray
            An array of bbox coords in pascal VOC format (x0, y0, x1, y1).
            Shape: (n_points, 4). Dtype: float64.
        points : np.ndarray
            The yx-coordinates of the non-zero points. Shape (n_points, 2). Dtype: int64
        scores : np.ndarray
            The probability values at the point coordinates. Shape (n_points,).
            Dtype: float32/float64.
        areas : np.ndarray
            The areas of the bounding boxes at the point coordinates. Shape (n_points,).
            Dtype: float32/float64.
        radius_outer : np.ndarray
            The radial distances to background at each point. Shape (n_points, )
        max_dist : float
            The maximum radial distance of all the radial distances
        score_threshold : float, default=0.5
            Threshold for the probability distance map.
        iou_threshold : float, default=0.5
            Threshold for the IoU metric deciding whether to suppres a bbox.

    Returns
    -------
        np.ndarray:
            The indices of the bboxes that are not suppressed. Shape: (n_kept, ).
    """
    keep = []

    if len(boxes) == 0:
        return np.zeros(0, dtype=np.int64)

    kdtree = KDTree(points, leafsize=16)

    suppressed = np.full(len(boxes), False)
    for current_idx in range(len(scores)):
        # If already visited or discarded
        if suppressed[current_idx]:
            continue

        # If score is already below threshold then break
        if scores[current_idx] < score_threshold:
            break

        # Query the points
        query = kdtree.query_ball_point(points[current_idx], max_dist)
        suppressed = _suppress_bbox(
            np.array(query), current_idx, boxes, areas, suppressed, iou_threshold
        )

        # Add the current box
        keep.append(current_idx)

    return np.array(keep)



from typing import Any, Dict, List, Optional

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from skimage.color import label2rgb
from tqdm import tqdm

try:
    import wandb
except ImportError:
    raise ImportError("wandb required. `pip install wandb`")

from ...inference import PostProcessor
from ...metrics.functional import iou_multiclass, panoptic_quality
from ...utils import get_type_instances, remap_label
from ..functional import iou

__all__ = ["WandbImageCallback", "WandbClassLineCallback", "WandbGetExamplesCallback"]


class WandbImageCallback(pl.Callback):
    def __init__(
        self,
        type_classes: Dict[str, int],
        sem_classes: Optional[Dict[str, int]],
        freq: int = 100,
    ) -> None:
        """Create a callback that logs prediction masks to wandb."""
        super().__init__()
        self.freq = freq
        self.type_classes = type_classes
        self.sem_classes = sem_classes

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Log the inputs and outputs of the model to wandb."""
        if batch_idx % self.freq == 0:
            outputs = outputs["soft_masks"]

            log_dict = {
                "global_step": trainer.global_step,
                "epoch": trainer.current_epoch,
            }

            img = batch["image"].detach().to("cpu").numpy()

            if "type" in list(batch.keys()):
                type_target = batch["type"].detach().to("cpu").numpy()
                soft_types = outputs["type"].detach().to("cpu")
                types = torch.argmax(F.softmax(soft_types, dim=1), dim=1).numpy()

                log_dict["val/cell_types"] = [
                    wandb.Image(
                        im.transpose(1, 2, 0),
                        masks={
                            "predictions": {
                                "mask_data": t,
                                "class_labels": self.type_classes,
                            },
                            "ground_truth": {
                                "mask_data": tt,
                                "class_labels": self.type_classes,
                            },
                        },
                    )
                    for im, t, tt in zip(img, types, type_target)
                ]

            if "sem" in list(batch.keys()):
                sem_target = batch["sem"].detach().to("cpu").numpy()
                soft_sem = outputs["sem"].detach().to(device="cpu")
                sem = torch.argmax(F.softmax(soft_sem, dim=1), dim=1).numpy()

                log_dict["val/tissue_areas"] = [
                    wandb.Image(
                        im.transpose(1, 2, 0),
                        masks={
                            "predictions": {
                                "mask_data": s,
                                "class_labels": self.sem_classes,
                            },
                            "ground_truth": {
                                "mask_data": st,
                                "class_labels": self.sem_classes,
                            },
                        },
                    )
                    for im, s, st in zip(img, sem, sem_target)
                ]

            for m in list(batch.keys()):
                if m not in ("sem", "type", "inst", "image"):
                    aux = outputs[m].detach().to(device="cpu")
                    log_dict[f"val/{m}"] = [
                        wandb.Image(a[i, ...], caption=f"{m} maps")
                        for a in aux
                        for i in range(a.shape[0])
                    ]

            trainer.logger.experiment.log(log_dict)


class WandbIoUCallback(pl.Callback):
    def __init__(
        self,
        type_classes: Dict[str, int],
        sem_classes: Optional[Dict[str, int]],
        freq: int = 100,
    ) -> None:
        """Create a base class for IoU wandb callbacks."""
        super().__init__()
        self.type_classes = type_classes
        self.sem_classes = sem_classes
        self.freq = freq

    def batch_end(self) -> None:
        """Abstract batch end method."""
        raise NotImplementedError

    def compute(
        self,
        key: str,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
    ) -> np.ndarray:
        """Compute the iou per class."""
        target = batch[key].detach()
        soft_types = outputs[key].detach()
        pred = F.softmax(soft_types, dim=1)

        met = iou(pred, target).mean(dim=0)
        return met.to("cpu").numpy()

    def train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Log the inputs and outputs of the model to wandb."""
        self.batch_end(trainer, outputs["soft_masks"], batch, batch_idx, phase="train")

    def validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Log the inputs and outputs of the model to wandb."""
        self.batch_end(trainer, outputs["soft_masks"], batch, batch_idx, phase="val")

    def test_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Log the inputs and outputs of the model to wandb."""
        self.batch_end(trainer, outputs["soft_masks"], batch, batch_idx, phase="test")


class WandbClassLineCallback(WandbIoUCallback):
    def __init__(
        self,
        type_classes: Dict[str, int],
        sem_classes: Optional[Dict[str, int]],
        freq: int = 100,
    ) -> None:
        """Create a wandb callback that logs per-class mIoU at batch ends."""
        super().__init__(type_classes, sem_classes, freq)

    def get_points(self, iou: np.ndarray, classes: Dict[int, str]) -> Any:
        """Return a wandb bar plot object of the current per class iou values."""
        return {lab: val for lab, val in zip(list(classes.values()), iou)}

    def batch_end(
        self,
        trainer: pl.Trainer,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        phase: str,
        dataloader_idx: int = None,
    ) -> None:
        """Log metrics at every 100th step to wandb."""
        if batch_idx % self.freq == 0:
            log_dict = {}
            if "type" in list(batch.keys()):
                iou = self.compute("type", outputs, batch)
                log_dict[f"{phase}/type_ious_points"] = self.get_points(
                    list(iou), self.type_classes
                )

            if "sem" in list(batch.keys()):
                iou = self.compute("sem", outputs, batch)
                log_dict[f"{phase}/sem_ious_points"] = self.get_points(
                    list(iou), self.sem_classes
                )

            trainer.logger.experiment.log(log_dict)

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Call the callback at val time."""
        self.validation_batch_end(
            trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        )

    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Call the callback at val time."""
        self.train_batch_end(
            trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        )


class WandbGetExamplesCallback(pl.Callback):
    def __init__(
        self,
        type_classes: Dict[str, int],
        sem_classes: Optional[Dict[str, int]],
        instance_postproc: str,
        inst_key: str,
        aux_key: str,
        inst_act: str = "softmax",
        aux_act: str = None,
    ) -> None:
        """Create a wandb callback that logs img examples at test time."""
        super().__init__()
        self.type_classes = type_classes
        self.sem_classes = sem_classes
        self.inst_key = inst_key
        self.aux_key = aux_key

        self.inst_act = inst_act
        self.aux_act = aux_act

        self.postprocessor = PostProcessor(
            instance_postproc=instance_postproc, inst_key=inst_key, aux_key=aux_key
        )

    def post_proc(
        self, outputs: Dict[str, torch.Tensor]
    ) -> List[Dict[str, np.ndarray]]:
        """Apply post processing to the outputs."""
        B, _, _, _ = outputs[self.inst_key].shape

        inst = outputs[self.inst_key].detach()
        if self.inst_act == "softmax":
            inst = F.softmax(inst, dim=1)
        if self.inst_act == "sigmoid":
            inst = torch.sigmoid(inst)

        aux = outputs[self.aux_key].detach()
        if self.aux_act == "tanh":
            aux = torch.tanh(aux)

        sem = None
        if "sem" in outputs.keys():
            sem = outputs["sem"].detach()
            sem = F.softmax(sem, dim=1).cpu().numpy()

        typ = None
        if "type" in outputs.keys():
            typ = outputs["type"].detach()
            typ = F.softmax(typ, dim=1).cpu().numpy()

        inst = inst.cpu().numpy()
        aux = aux.cpu().numpy()
        outmaps = []
        for i in range(B):
            maps = {
                self.inst_key: inst[i],
                self.aux_key: aux[i],
            }
            if sem is not None:
                maps["sem"] = sem[i]
            if typ is not None:
                maps["type"] = typ[i]

            out = self.postprocessor.post_proc_pipeline(maps)
            outmaps.append(out)

        return outmaps

    def count_pixels(self, img: np.ndarray, shape: int):
        """Compute pixel proportions per class."""
        return [float(p) / shape**2 for p in np.bincount(img.astype(int).flatten())]

    def epoch_end(self, trainer, pl_module) -> None:
        """Log metrics at the epoch end."""
        decs = [list(it.keys()) for it in pl_module.heads.values()]
        outheads = [item for sublist in decs for item in sublist]

        loader = trainer.datamodule.test_dataloader()
        runid = trainer.logger.experiment.id
        test_res_at = wandb.Artifact("test_pred_" + runid, "test_preds")

        # Create artifact
        runid = trainer.logger.experiment.id
        test_res_at = wandb.Artifact("test_pred_" + runid, "test_preds")

        # Init wb table
        cols = ["id", "inst_gt", "inst_pred", "bPQ"]

        if "type" in outheads:
            cols += [
                "cell_types",
                *[f"pq_{c}" for c in self.type_classes.values() if c != "bg"],
            ]
        if "sem" in outheads:
            cols += [
                "tissue_types",
                *[f"iou_{c}" for c in self.sem_classes.values() if c != "bg"],
            ]

        model_res_table = wandb.Table(columns=cols)

        #
        with tqdm(loader, unit="batch") as loader:
            with torch.no_grad():
                for batch_idx, batch in enumerate(loader):
                    soft_masks = pl_module.forward(batch["image"].to(pl_module.device))
                    imgs = list(batch["image"].detach().cpu().numpy())  # [(C, H, W)*B]
                    inst_targets = list(batch["inst_map"].detach().cpu().numpy())

                    outmaps = self.post_proc(soft_masks)

                    type_targets = None
                    if "type" in list(batch.keys()):
                        type_targets = list(
                            batch["type"].detach().cpu().numpy()
                        )  # [(C, H, W)*B]

                    sem_targets = None
                    if "sem" in list(batch.keys()):
                        sem_targets = list(
                            batch["sem"].detach().cpu().numpy()
                        )  # [(C, H, W)*B]

                    # loop the images in batch
                    for i, (pred, im, inst_target) in enumerate(
                        zip(outmaps, imgs, inst_targets)
                    ):
                        inst_targ = remap_label(inst_target)
                        inst_pred = remap_label(pred["inst"])

                        wb_inst_gt = wandb.Image(label2rgb(inst_targ, bg_label=0))
                        wb_inst_pred = wandb.Image(label2rgb(inst_pred, bg_label=0))
                        pq_inst = panoptic_quality(inst_targ, inst_pred)["pq"]

                        row = [
                            f"test_batch_{batch_idx}_{i}",
                            wb_inst_gt,
                            wb_inst_pred,
                            pq_inst,
                        ]

                        if type_targets is not None:
                            per_class_pq = [
                                panoptic_quality(
                                    remap_label(
                                        get_type_instances(
                                            inst_targ, type_targets[i], j
                                        )
                                    ),
                                    remap_label(
                                        get_type_instances(inst_pred, pred["type"], j)
                                    ),
                                )["pq"]
                                for j in self.type_classes.keys()
                                if j != 0
                            ]

                            type_classes_set = wandb.Classes(
                                [
                                    {"name": name, "id": id}
                                    for id, name in self.type_classes.items()
                                    if id != 0
                                ]
                            )
                            wb_type = wandb.Image(
                                im.transpose(1, 2, 0),
                                classes=type_classes_set,
                                masks={
                                    "ground_truth": {"mask_data": type_targets[i]},
                                    "pred": {"mask_data": pred["type"]},
                                },
                            )

                            row += [wb_type, *per_class_pq]

                        if sem_targets is not None:
                            per_class_iou = list(
                                iou_multiclass(
                                    sem_targets[i], pred["sem"], len(self.sem_classes)
                                )
                            )

                            sem_classes_set = wandb.Classes(
                                [
                                    {"name": name, "id": id}
                                    for id, name in self.sem_classes.items()
                                    if id != 0
                                ]
                            )
                            wb_sem = wandb.Image(
                                im.transpose(1, 2, 0),
                                classes=sem_classes_set,
                                masks={
                                    "ground_truth": {"mask_data": sem_targets[i]},
                                    "pred": {"mask_data": pred["sem"]},
                                },
                            )
                            row += [wb_sem, *per_class_iou[1:]]

                        model_res_table.add_data(*row)

        test_res_at.add(model_res_table, "model_batch_results")
        trainer.logger.experiment.log_artifact(test_res_at)

    def on_test_epoch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
    ) -> None:
        """Call the callback at test time."""
        self.epoch_end(trainer, pl_module)



import torch
import torch.nn as nn

from .criterions import SEG_LOSS_LOOKUP


class Loss(nn.Module):
    def __init__(
        self,
        name: str,
        apply_sd: bool = False,
        apply_ls: bool = False,
        apply_svls: bool = False,
        edge_weight: float = None,
        class_weights: torch.Tensor = None,
        **kwargs,
    ) -> None:
        """Loss wrapper class.

        Parameters:
        -----------
            name : str
                Name of the loss function.
            apply_sd : bool, default=False
                If True, Spectral decoupling regularization will be applied  to the
                loss matrix.
            apply_ls : bool, default=False
                If True, Label smoothing will be applied to the target.
            apply_svls : bool, default=False
                If True, spatially varying label smoothing will be applied to the target
            edge_weight : float, default=none
                Weight that is added to object borders.
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (n_classes,).

        Raises
        ------
            ValueError: if the loss function name is illegal.
        """
        super().__init__()

        allowed = list(SEG_LOSS_LOOKUP.keys())
        if name not in allowed:
            raise ValueError(
                f"Illegal loss function given. Allowed: {allowed}. Got: '{name}'"
            )

        self.loss = SEG_LOSS_LOOKUP[name](
            apply_sd=apply_sd,
            apply_ls=apply_ls,
            apply_svls=apply_svls,
            edge_weight=edge_weight,
            class_weights=class_weights,
            **kwargs,
        )

    def forward(
        self,
        yhat: torch.Tensor,
        target: torch.Tensor,
        target_weight: torch.Tensor = None,
        **kwargs,
    ) -> torch.Tensor:
        """Forward pass for the loss function."""
        return self.loss(
            yhat=yhat, target=target, target_weight=target_weight, **kwargs
        )



import numpy as np
import pytest

from cellseg_models_pytorch.postproc import (
    fill_holes_and_remove_small_masks,
    gen_flows,
    post_proc_cellpose,
    post_proc_cellpose_old,
    post_proc_dcan,
    post_proc_dran,
    post_proc_drfns,
    post_proc_hovernet,
    post_proc_omnipose,
    post_proc_stardist,
)
from cellseg_models_pytorch.transforms.functional import (
    gen_contour_maps,
    gen_dist_maps,
    gen_flow_maps,
    gen_hv_maps,
    gen_stardist_maps,
)
from cellseg_models_pytorch.utils import FileHandler, binarize, remap_label


def test_gen_flows(inst_map):
    flows = gen_flow_maps(inst_map)
    flows = gen_flows(flows)

    assert flows.shape == (inst_map.shape[0], inst_map.shape[1], 3)


def test_fill_holes(inst_map):
    out = fill_holes_and_remove_small_masks(inst_map.copy(), min_size=7)

    assert len(np.unique(out)) == 6


@pytest.mark.parametrize("return_flows", [True, False])
@pytest.mark.parametrize("interp", [True, False])
def test_postproc_cellpose(inst_map, return_flows, interp):
    flows = gen_flow_maps(inst_map)

    if not return_flows:
        rebuild = post_proc_cellpose(
            inst_map, flows, min_size=0, interp=interp, use_gpu=False
        )
    else:
        rebuild, _ = post_proc_cellpose(
            inst_map,
            flows,
            min_size=0,
            return_flows=return_flows,
            interp=interp,
            use_gpu=False,
        )

    assert rebuild.shape == inst_map.shape
    assert rebuild.dtype == "int32"

    if interp:
        assert len(np.unique(rebuild)) == 4


@pytest.mark.parametrize("return_flows", [True, False])
def test_postproc_omnipose(inst_map, return_flows):
    flows = gen_flow_maps(inst_map)

    if not return_flows:
        rebuild = post_proc_omnipose(inst_map, flows, min_size=0)
    else:
        rebuild, _ = post_proc_omnipose(
            inst_map, flows, min_size=0, return_flows=return_flows
        )

    assert rebuild.dtype == "int32"
    np.testing.assert_array_equal(
        np.unique(remap_label(inst_map)), np.unique(remap_label(rebuild))
    )


def test_postproc_cellpose_old(inst_map):
    flows = gen_hv_maps(inst_map)
    rebuild = rebuild = post_proc_cellpose_old(inst_map, flows)

    assert rebuild.shape == inst_map.shape
    assert rebuild.dtype == "int32"


@pytest.mark.parametrize("enhance", [True, False])
def test_postproc_hovernet(inst_map, enhance):
    hover = gen_hv_maps(inst_map)
    rebuild = post_proc_hovernet(binarize(inst_map), hover, enhance=enhance)

    assert rebuild.shape == inst_map.shape
    assert rebuild.dtype == "int32"


def test_postproc_drfns(inst_map):
    dist = gen_dist_maps(inst_map)
    rebuild = post_proc_drfns(inst_map, dist)

    assert rebuild.dtype == "int32"
    assert rebuild.shape == inst_map.shape
    np.testing.assert_array_equal(
        np.unique(remap_label(inst_map)), np.unique(remap_label(rebuild))
    )


def test_postproc_dcan(inst_map):
    cont = gen_contour_maps(inst_map)
    rebuild = post_proc_dcan(inst_map, cont)

    assert rebuild.dtype == "int32"
    assert rebuild.shape == inst_map.shape


def test_postproc_dran(inst_map):
    cont = gen_contour_maps(inst_map)
    rebuild = post_proc_dran(inst_map, cont)

    assert rebuild.dtype == "int32"
    assert rebuild.shape == inst_map.shape


def test_postproc_stardist(mask_patch_dir):
    mask_path = sorted(mask_patch_dir.glob("*"))[0]
    inst_map = FileHandler.read_mat(mask_path)
    stardist = gen_stardist_maps(inst_map, 32)
    dist = gen_dist_maps(inst_map)
    rebuild = post_proc_stardist(dist, stardist)

    assert rebuild.dtype == "int32"
    assert rebuild.shape == inst_map.shape



from collections import OrderedDict
from pathlib import Path
from typing import Dict, List

import numpy as np

from ..utils import FileHandler
from ._base_dataset import TrainDatasetBase

IMG_SUFFIXES = (".jpeg", ".jpg", ".tif", ".tiff", ".png")
MASK_SUFFIXES = (".mat",)

__all__ = ["SegmentationFolderDataset"]


class SegmentationFolderDataset(TrainDatasetBase):
    def __init__(
        self,
        path: str,
        mask_path: str,
        img_transforms: List[str],
        inst_transforms: List[str],
        normalization: str = None,
        return_inst: bool = False,
        return_binary: bool = True,
        return_type: bool = True,
        return_sem: bool = False,
        return_weight: bool = False,
        **kwargs,
    ) -> None:
        """Create a dataset class that reads images/patches from a folder.

        Parameters
        ----------
            path : str
                Path to the folder containing the images.
            mask_path : str
                Path to the folder containing the corresponding masks (.mat files).
            img_transforms : List[str]
                A list containing all the transformations that are applied to the input
                images and corresponding masks. Allowed ones: "blur", "non_spatial",
                "non_rigid", "rigid", "hue_sat", "random_crop", "center_crop", "resize"
            inst_transforms : List[str]
                A list containg all the transformations that are applied to only the
                instance labelled masks. Allowed ones: "cellpose", "contour", "dist",
                "edgeweight", "hovernet", "omnipose", "smooth_dist", "binarize"
            normalization : str, optional
                Apply img normalization after all the transformations. One of "minmax",
                "norm", "percentile", None.
            return_inst : bool, default=False
                If True, returns the instance labelled mask. (If the db contains these.)
            return_binary : bool, default=True
                If True, returns a binarized instance mask. (If the db contains these.)
            return_type : bool, default=True
                If True, returns a type mask. (If the db contains these.)
            return_sem : bool, default=False
                If True, returns a semantic mask, (If the db contains these.)
            return_weight : bool, default=False
                Include a nuclear border weight map in the output.

        Raises
        ------
            ValueError if there are issues with the given paths or files.
        """
        super().__init__(
            img_transforms=img_transforms,
            inst_transforms=inst_transforms,
            normalization=normalization,
            return_inst=return_inst,
            return_type=return_type,
            return_sem=return_sem,
            return_weight=return_weight,
            return_binary=return_binary,
            **kwargs,
        )

        self.path = Path(path)
        self.mask_path = Path(mask_path)

        if not self.path.exists():
            raise ValueError(f"folder: {path} does not exist")

        if not self.path.is_dir():
            raise ValueError(f"path: {path} is not a folder")

        if not all([f.suffix in IMG_SUFFIXES for f in self.path.iterdir()]):
            raise ValueError(
                f"files formats in given folder need to be in {IMG_SUFFIXES}"
            )

        if not self.mask_path.exists():
            raise ValueError(f"folder: {self.mask_path} does not exist")

        if not self.mask_path.is_dir():
            raise ValueError(f"path: {self.mask_path} is not a folder")

        if not all([f.suffix in MASK_SUFFIXES for f in self.mask_path.iterdir()]):
            raise ValueError(
                f"files formats in given folder need to be in {MASK_SUFFIXES}"
            )

        self.fnames_imgs = sorted(self.path.glob("*"))
        self.fnames_masks = sorted(self.mask_path.glob("*"))
        if len(self.fnames_imgs) != len(self.fnames_masks):
            raise ValueError(
                f"Found different number of files in {self.path.as_posix()} and "
                f"{self.mask_path.as_posix()}."
            )

    def read_img_mask(
        self,
        ix: int,
        return_type: bool = True,
        return_sem: bool = False,
    ) -> Dict[str, np.ndarray]:
        """Read img & mask patches at index `ix`.

        Parameters
        ----------
            ix : int
                Index for the img in the folder.
            return_type : bool, default=True
                If True, returns a type mask. (If the db contains these.)
            return_sem : bool, default=False
                If True, returns a semantic mask, (If the db contains these.)

        Returns
        -------
            OrderedDict[str, np.ndarray]:
                A Dict of numpy matrices. Img shape: (H, W, 3), mask shapes: (H, W).
                keys of the dict are: "im", "inst", "type", "sem"

        Raises
        ------
            KeyError: If a mask that does not exist in a given .mat file.
        """
        out = OrderedDict()
        out["image"] = FileHandler.read_img(self.fnames_imgs[ix])
        masks = FileHandler.read_mat(self.fnames_masks[ix], return_all=True)

        try:
            out["inst"] = masks["inst_map"]
        except KeyError:
            raise KeyError(
                f"The file {self.fnames_masks[ix]} does not contain key `inst_map`. "
                "Try setting `return_inst=False`."
            )

        if return_type:
            try:
                out["type"] = masks["type_map"]
            except KeyError:
                raise KeyError(
                    f"The file {self.fnames_masks[ix]} does not contain key `type_map`."
                    " Try setting `return_type=False`."
                )

        if return_sem:
            try:
                out["sem"] = masks["sem_map"]
            except KeyError:
                raise KeyError(
                    f"The file {self.fnames_masks[ix]} does not contain key `sem_map`."
                    "Try setting `return_sem=False`."
                )

        return out

    def __len__(self) -> int:
        """Return the number of items in the db."""
        return len(self.fnames_imgs)

    def __getitem__(self, ix: int) -> Dict[str, np.ndarray]:
        """Get item.

        Parameters
        ----------
            ix : int
                An index for the iterable dataset.

        Returns
        -------
            Dict[str, np.ndarray]:
                A dictionary containing all the augmented data patches.
                Keys are: "im", "inst", "type", "sem". Image shape: (B, 3, H, W).
                Mask shapes: (B, C_mask, H, W).

        """
        return self._getitem(ix, self.read_img_mask)



from collections import OrderedDict
from pathlib import Path
from typing import Dict, List

import numpy as np

from ._base_dataset import TrainDatasetBase

try:
    import tables as tb
except Exception:
    raise ImportError(
        "`pytables` needed for this class. Install with: `pip install tables`"
    )


__all__ = ["SegmentationHDF5Dataset"]


class SegmentationHDF5Dataset(TrainDatasetBase):
    def __init__(
        self,
        path: str,
        img_transforms: List[str],
        inst_transforms: List[str],
        normalization: str = None,
        return_inst: bool = False,
        return_binary: bool = True,
        return_type: bool = True,
        return_sem: bool = False,
        return_weight: bool = False,
        **kwargs,
    ) -> None:
        """Create a dataset class that reads images/patches from a hdf5 database.

        NOTE: The h5-db needs to be written with `HDF5Writer` (/utils/h5_writer.py).

        Parameters
        ----------
            path : str
                Path to the hdf5 db.
            img_transforms : List[str]
                A list containing all the transformations that are applied to the input
                images and corresponding masks. Allowed ones: "blur", "non_spatial",
                "non_rigid", "rigid", "hue_sat", "random_crop", "center_crop", "resize"
            inst_transforms : List[str]
                A list containg all the transformations that are applied to only the
                instance labelled masks. Allowed ones: "cellpose", "contour", "dist",
                "edgeweight", "hovernet", "omnipose", "smooth_dist", "binarize"
            normalization : str, optional
                Apply img normalization after all the transformations. One of "minmax",
                "norm", "percentile", None.
            return_inst : bool, default=False
                If True, returns the instance labelled mask. (If the db contains these.)
            return_binary : bool, default=True
                If True, returns a binarized instance mask. (If the db contains these.)
            return_type : bool, default=True
                If True, returns a type mask. (If the db contains these.)
            return_sem : bool, default=False
                If True, returns a semantic mask, (If the db contains these.)
            return_weight : bool, default=False
                Include a nuclear border weight map in the output.
        """
        super().__init__(
            img_transforms=img_transforms,
            inst_transforms=inst_transforms,
            normalization=normalization,
            return_inst=return_inst,
            return_type=return_type,
            return_sem=return_sem,
            return_weight=return_weight,
            return_binary=return_binary,
        )

        self.path = Path(path)

        if self.path.suffix not in (".h5", ".hdf5"):
            raise ValueError(
                f"""The input path has to be a hdf5 db. Got suffix: {self.path.suffix}
                Allowed suffices: {(".h5", ".hdf5")}"""
            )

        # n imgs in the db.
        with tb.open_file(self.path) as h5:
            self.n_items = h5.root._v_attrs["n_items"]

    def read_h5_patch(
        self,
        ix: int,
        return_type: bool = True,
        return_sem: bool = False,
    ) -> Dict[str, np.ndarray]:
        """Read img & mask patches at index `ix`.

        Parameters
        ----------
            ix : int
                Index for the hdf5 db-arrays.
            return_type : bool, default=True
                If True, returns a type mask. (If the db contains these.)
            return_sem : bool, default=False
                If True, returns a semantic mask, (If the db contains these.)

        Returns
        -------
            OrderedDict[str, np.ndarray]:
                A Dict of numpy matrices. Img shape: (H, W, 3), mask shapes: (H, W).
                keys of the dict are: "im", "inst", "type", "sem"

        Raises
        ------
            IOError: If a mask that does not exist in the db is being read.
        """
        with tb.open_file(self.path.as_posix(), "r") as h5:
            out = OrderedDict()
            out["image"] = h5.root.imgs[ix, ...]

            try:
                out["inst"] = h5.root.insts[ix, ...]
            except Exception:
                raise IOError(
                    "The HDF5 database does not contain instance labelled masks. Try "
                    "setting `return_inst=False`"
                )

            if return_type:
                try:
                    out["type"] = h5.root.types[ix, ...]
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain type masks. Try setting "
                        "`return_type = False` "
                    )

            if return_sem:
                try:
                    out["sem"] = h5.root.areas[ix, ...]
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain semantic masks. Try "
                        "setting `return_sem = False`"
                    )

            return out

    def __len__(self) -> int:
        """Return the number of items in the db."""
        return self.n_items

    def __getitem__(self, ix: int) -> Dict[str, np.ndarray]:
        """Get item.

        Parameters
        ----------
            ix : int
                An index for the iterable dataset.

        Returns
        -------
            Dict[str, np.ndarray]:
                A dictionary containing all the augmented data patches.
                Keys are: "im", "inst", "type", "sem". Image shape: (B, 3, H, W).
                Mask shapes: (B, C_mask, H, W).

        """
        return self._getitem(ix, self.read_h5_patch)



from torch.nn import (
    CELU,
    ELU,
    GELU,
    GLU,
    SELU,
    Hardshrink,
    Hardsigmoid,
    LeakyReLU,
    PReLU,
    ReLU,
    ReLU6,
    RReLU,
    Sigmoid,
    SiLU,
    Tanh,
    Tanhshrink,
)

from .gated_gelu import GEGLU, ApproximateGELU
from .mish import Mish
from .star_relu import StarReLU
from .swish import Swish

ACT_LOOKUP = {
    "mish": Mish,
    "swish": Swish,
    "relu": ReLU,
    "relu6": ReLU6,
    "rrelu": RReLU,
    "selu": SELU,
    "celu": CELU,
    "gelu": GELU,
    "glu": GLU,
    "tanh": Tanh,
    "sigmoid": Sigmoid,
    "silu": SiLU,
    "prelu": PReLU,
    "leaky-relu": LeakyReLU,
    "elu": ELU,
    "hardshrink": Hardshrink,
    "tanhshrink": Tanhshrink,
    "hardsigmoid": Hardsigmoid,
    "star_relu": StarReLU,
    "geglu": GEGLU,
    "approximate_geglu": ApproximateGELU,
}

__all__ = [
    "Mish",
    "Swish",
    "ReLU",
    "ReLU6",
    "RReLU",
    "SELU",
    "CELU",
    "GELU",
    "GLU",
    "Tanh",
    "Sigmoid",
    "SiLU",
    "PReLU",
    "LeakyReLU",
    "ELU",
    "Hardshrink",
    "Tanhshrink",
    "Hardsigmoid",
    "Activation",
    "ACT_LOOKUP",
    "GEGLU",
    "ApproximateGELU",
]



"""
Adapted from the cellpose repo.

https://github.com/MouseLand/cellpose

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. Redistributions in
binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution. Neither the name of HHMI
nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
“AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
import scipy.ndimage as ndi
import torch

from .cellpose import normalize_field

__all__ = ["smooth_distance", "gen_omni_flow_maps"]


def smooth_distance(
    inst_map: np.ndarray, device: str = "cpu", pad: int = 1
) -> np.ndarray:
    """Smooth FIM (Fast Iterative Method) based distance transform.

    Avoids boundary pixellation that can cause artefacts in flows.

    https://www.biorxiv.org/content/10.1101/2021.11.03.467199v2

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        device str : default="cpu"
            One of "cuda" or "cpu".
        pad : int, default=1
            The number of pixels padded around the input.

    Returns
    -------
        np.ndarray:
            Smooth distance transform. Shape: (H, W).
    """
    dists = ndi.distance_transform_edt(inst_map)
    inst_map = np.pad(inst_map, pad)
    H, W = inst_map.shape

    # get non-zero pixel indices
    y, x = np.nonzero(inst_map)

    # 9-pixel neighborhood for every nonzero pixel
    yneighbors = np.stack((y - 1, y - 1, y - 1, y, y, y, y + 1, y + 1, y + 1), axis=0)
    xneighbors = np.stack((x - 1, x, x + 1, x - 1, x, x + 1, x - 1, x, x + 1), axis=0)

    # get neighbor validator (not all neighbors are in same mask)
    # extract list of label values,
    neighbor_masks = inst_map[yneighbors, xneighbors]
    isneighbor = neighbor_masks == neighbor_masks[4]  # 4 corresponds to x, y now

    # set number of iterations
    n_iter = np.ceil(np.max(dists) * 1.16).astype(int) + 1
    nimg = xneighbors.shape[0] // 9

    # solve the eikonal heat eq
    pty = torch.from_numpy(yneighbors).to(device)
    ptx = torch.from_numpy(xneighbors).to(device)
    T = torch.zeros((nimg, H, W), dtype=torch.double, device=device)
    isneigh = torch.from_numpy(isneighbor).to(device)

    for _ in range(n_iter):
        Tneigh = T[:, pty, ptx]
        Tneigh *= isneigh

        # using flattened index for the lattice points, just like gradient below
        minx = torch.minimum(Tneigh[:, 3, :], Tneigh[:, 5, :])
        mina = torch.minimum(Tneigh[:, 2, :], Tneigh[:, 6, :])
        miny = torch.minimum(Tneigh[:, 1, :], Tneigh[:, 7, :])
        minb = torch.minimum(Tneigh[:, 0, :], Tneigh[:, 8, :])

        A = torch.where(
            torch.abs(mina - minb) >= 2,
            torch.minimum(mina, minb) + np.sqrt(2),
            0.5 * (mina + minb + torch.sqrt(4 - (mina - minb) ** 2)),
        )
        B = torch.where(
            torch.abs(miny - minx) >= np.sqrt(2),
            torch.minimum(miny, minx) + 1,
            0.5 * (miny + minx + torch.sqrt(2 - (miny - minx) ** 2)),
        )

        T[:, pty[4, ...], ptx[4, ...]] = torch.sqrt(A * B + 1e-12)

    return T.cpu().squeeze().numpy()[pad:-pad, pad:-pad]


# Adapted from
# https://github.com/MouseLand/cellpose/blob/master/cellpose/omnipose/core.py
def gen_omni_flow_maps(inst_map: np.ndarray, pad: int = 1) -> np.ndarray:
    """Generate eikonal flow maps from inst maps like in OmniPose.

    I.e. take the first order x- and -y derivatives of eikonal distance
    transfroms of the cell masks.

    Omnipose: a high-precision morphology-independent solution for
    bacterial cell segmentation Kevin J. Cutler, Carsen Stringer,
    Paul A. Wiggins, Joseph D. Mougous bioRxiv 2021.11.03.467199;
    doi: https://doi.org/10.1101/2021.11.03.467199

    https://www.biorxiv.org/content/10.1101/2021.11.03.467199v2

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        pad (int, default=1):
            number of pixels for constant padding

    Returns
    -------
        np.ndarray:
            Y and X- flows in this order. Shape (2, H, W). Dtype: float64.
    """
    # FMI euclidean distance transform
    dists = smooth_distance(inst_map)

    # number of iterations for heat diffusion equation
    niter = np.ceil(np.max(dists) * 1.16).astype(int) + 1

    H, W = inst_map.shape
    mu = np.zeros((2, H, W), np.float64)

    inst_list = list(np.unique(inst_map))
    if 0 in inst_list:
        inst_list.remove(0)

    for inst_id in inst_list:
        inst = np.array(inst_map == inst_id, np.int32)

        # padded bounding box of the instance
        sy, sx = ndi.find_objects(inst)[0]
        inst = np.pad(inst[sy, sx], pad)
        h, w = inst.shape

        # non-zero pixel indices in the bounding box
        y, x = np.nonzero(inst)

        # solve heat equation for eikonal distance
        T = np.zeros(h * w, dtype=np.float64)
        for _ in range(niter):
            minx = np.minimum(T[y * w + x - 1], T[y * w + x + 1])
            miny = np.minimum(
                T[(y - 1) * w + x],
                T[(y + 1) * w + x],
            )
            mina = np.minimum(T[(y - 1) * w + x - 1], T[(y + 1) * w + x + 1])
            minb = np.minimum(T[(y - 1) * w + x + 1], T[(y + 1) * w + x - 1])

            a = 4 - (mina - minb) ** 2
            a[a < 0] = 0.0
            A = np.where(
                np.abs(mina - minb) >= 2,
                np.minimum(mina, minb) + np.sqrt(2),
                0.5 * (mina + minb + np.sqrt(a)),
            )

            b = 2 - (miny - minx) ** 2
            b[b < 0] = 0.0
            B = np.where(
                np.abs(miny - minx) >= np.sqrt(2),
                np.minimum(miny, minx) + 1,
                0.5 * (miny + minx + np.sqrt(b)),
            )

            T[y * w + x] = np.sqrt(A * B + 1e-12)

        # central difference approximation to first derivative
        dy = (T[(y + 1) * w + x] - T[(y - 1) * w + x]) / 2
        dx = (T[y * w + x + 1] - T[y * w + x - 1]) / 2

        mu[:, (sy.start + y - pad), (sx.start + x - pad)] = np.stack([dy, dx])

    mu = normalize_field(mu)

    return mu



"""
Adapted from the cellpose repo.

https://github.com/MouseLand/cellpose

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. Redistributions in
binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution. Neither the name of HHMI
nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
“AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
import scipy.ndimage as ndi
from numba import njit

__all__ = ["normalize_field", "gen_flow_maps"]


def normalize_field(mu: np.ndarray) -> np.ndarray:
    """Normalize the flow field.

    Parameters
    ----------
        mu : np.ndarray
            The un-normalized y- and x- flows. Shape (2, H, W).

    Returns
    -------
        np.ndarray:
            The normalized y- and x- flows. Shape (2, H, W).
    """
    mag = np.sqrt(np.nansum(mu**2, axis=0))
    mask = np.logical_and(mag != 0, ~np.isnan(mag))
    out = np.zeros_like(mu)
    out = np.divide(mu, mag, out=out, where=mask)

    return out


@njit(nogil=True)
def _extend_centers(
    T: np.ndarray,
    y: np.ndarray,
    x: np.ndarray,
    ymed: int,
    xmed: int,
    w: int,
    niter: int,
) -> np.ndarray:
    """Run diffusion from center of the label."""
    for _ in range(niter):
        T[ymed * w + xmed] += 1

        t = T[y * w + x]
        tdy = T[(y - 1) * w + x] + T[(y + 1) * w + x]
        tdx = T[y * w + x - 1] + T[y * w + x + 1]
        tdydx = T[(y - 1) * w + x - 1] + T[(y - 1) * w + x + 1]
        tdxdy = T[(y + 1) * w + x - 1] + T[(y + 1) * w + x + 1]

        T[y * w + x] = 1 / 9.0 * (t + tdy + tdx + tdydx + tdxdy)

    return T


def gen_flow_maps(inst_map: np.ndarray, pad: int = 1) -> np.ndarray:
    """Generate flow maps from inst maps like in CellPose.

    I.e. take the x- and y- derivates from a time-independent heat
    diffusion map of the instances.

    Stringer, C., Wang, T., Michaelos, M. et al. Cellpose:
    a generalist algorithm for cellular segmentation. Nat Methods 18,
    100-106 (2021). https://doi.org/10.1038/s41592-020-01018-x

    https://www.nature.com/articles/s41592-020-01018-x

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        pad : int, default=1
            Number of pixels for constant padding.

    Returns
    -------
        np.ndarray:
            Y and X- flows in this order. Shape (2, H, W). Dtype: float64
    """
    H, W = inst_map.shape
    mu = np.zeros((2, H, W), np.float64)
    mu_c = np.zeros((H, W), np.float64)

    inst_list = list(np.unique(inst_map))
    if 0 in inst_list:
        inst_list.remove(0)

    # loop the instances
    for inst_id in inst_list:
        inst = np.array(inst_map == inst_id, np.int32)

        # padded bounding box of the instance
        sy, sx = ndi.find_objects(inst)[0]
        inst = np.pad(inst[sy, sx], pad)
        h, w = inst.shape

        # non-zero pixel indices in the bounding box
        y, x = np.nonzero(inst)

        # number of iterations for heat diffusion equation
        niter = 2 * np.int32((np.ptp(x) + np.ptp(y)))

        # center of mass for the instance
        ymed = np.median(y)
        xmed = np.median(x)
        imin = np.argmin((x - xmed) ** 2 + (y - ymed) ** 2)
        xmed = np.array([x[imin]], np.int32)
        ymed = np.array([y[imin]], np.int32)

        # solve heat equation
        T = np.zeros(h * w, dtype=np.float64)
        T = _extend_centers(T, y, x, ymed, xmed, np.int32(w), np.int32(niter))
        T[(y + 1) * w + x + 1] = np.log(1.0 + T[(y + 1) * w + x + 1])

        # central difference approximation to first derivative
        dy = (T[(y + 1) * w + x] - T[(y - 1) * w + x]) / 2
        dx = (T[y * w + x + 1] - T[y * w + x - 1]) / 2

        mu[:, (sy.start + y - pad), (sx.start + x - pad)] = np.stack([dy, dx])
        mu_c[sy.start + y - pad, sx.start + x - pad] = T[y * w + x]  # heat diff

    mu = normalize_field(mu)

    return mu



import numpy as np
import pytest
import torch

from cellseg_models_pytorch.transforms.albu_transforms import (
    apply_each,
    blur_transforms,
    cellpose_transform,
    compose,
    edgeweight_transform,
    non_rigid_transforms,
    smooth_dist_transform,
    to_tensorv3,
)


@pytest.mark.parametrize("zero_input", [None, np.zeros((11, 10))])
def test_full_transform_pipeline(img_sample, inst_map, sem_map, zero_input):
    basic_transforms = compose([blur_transforms(), non_rigid_transforms()])
    inst_transforms = apply_each(
        [edgeweight_transform(), cellpose_transform(), smooth_dist_transform()]
    )
    to_tensor = to_tensorv3()

    if zero_input is not None:
        inst_map = zero_input

    aug = basic_transforms(image=img_sample, masks=[inst_map, sem_map])
    aux = inst_transforms(image=aug["image"], inst_map=aug["masks"][0])
    all_data = to_tensor(image=aug["image"], masks=aug["masks"], aux=aux)

    assert all_data["image"].shape == torch.Size([3, 500, 500])
    assert all_data["image"].dtype == torch.float32
    assert all_data["aux"]["edgeweight"].shape == torch.Size([11, 10])
    assert all_data["aux"]["cellpose"].shape == torch.Size([2, 11, 10])
    assert all_data["aux"]["smoothdist"].shape == torch.Size([11, 10])
    assert all(mask.shape == torch.Size([11, 10]) for mask in all_data["masks"])



import torch
import torch.nn.functional as F

from cellseg_models_pytorch.utils import tensor_one_hot

from ..weighted_base_loss import WeightedBaseLoss


class TverskyLoss(WeightedBaseLoss):
    def __init__(
        self,
        alpha: float = 0.7,
        beta: float = 0.3,
        apply_sd: bool = False,
        apply_ls: bool = False,
        apply_svls: bool = False,
        edge_weight: float = None,
        class_weights: torch.Tensor = None,
        **kwargs
    ) -> None:
        """Tversky loss.

        https://arxiv.org/abs/1706.05721

        Parameters
        ----------
            alpha : float, default=0.7
                 False positive dice coefficient.
            beta : float, default=0.3
                False negative tanimoto coefficient.
            apply_sd : bool, default=False
                If True, Spectral decoupling regularization will be applied  to the
                loss matrix.
            apply_ls : bool, default=False
                If True, Label smoothing will be applied to the target.
            apply_svls : bool, default=False
                If True, spatially varying label smoothing will be applied to the target
            edge_weight : float, default=none
                Weight that is added to object borders.
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (n_classes,).
        """
        super().__init__(apply_sd, apply_ls, apply_svls, class_weights, edge_weight)
        self.alpha = alpha
        self.beta = beta
        self.eps = 1e-7

    def forward(
        self,
        yhat: torch.Tensor,
        target: torch.Tensor,
        target_weight: torch.Tensor = None,
        **kwargs
    ) -> torch.Tensor:
        """Compute the Tversky loss.

        Parameters
        ----------
            yhat : torch.Tensor
                The prediction map. Shape (B, C, H, W).
            target : torch.Tensor
                the ground truth annotations. Shape (B, H, W).
            target_weight : torch.Tensor, default=None
                The edge weight map. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed Tversky loss (scalar).
        """
        num_classes = yhat.shape[1]
        target_one_hot = tensor_one_hot(target, n_classes=num_classes)
        yhat_soft = F.softmax(yhat, dim=1)
        assert target_one_hot.shape == yhat.shape

        if self.apply_svls:
            target_one_hot = self.apply_svls_to_target(
                target_one_hot, num_classes, **kwargs
            )

        if self.apply_ls:
            target_one_hot = self.apply_ls_to_target(
                target_one_hot, num_classes, **kwargs
            )

        intersection = torch.sum(yhat_soft * target_one_hot, 1)
        fps = torch.sum(yhat_soft * (1.0 - target_one_hot), 1)
        fns = torch.sum(target_one_hot * (1.0 - yhat_soft), 1)
        denom = intersection + self.alpha * fps + self.beta * fns
        tversky_loss = intersection / denom.clamp_min(self.eps)

        if self.apply_sd:
            tversky_loss = self.apply_spectral_decouple(tversky_loss, yhat)

        if self.class_weights is not None:
            tversky_loss = self.apply_class_weights(tversky_loss, target)

        if self.edge_weight is not None:
            tversky_loss = self.apply_edge_weights(tversky_loss, target_weight)

        return torch.mean(1.0 - tversky_loss)



import pytest
import torch

from cellseg_models_pytorch.transforms.albu_transforms import (
    blur_transforms,
    center_crop,
    compose,
    hue_saturation_transforms,
    non_rigid_transforms,
    non_spatial_transforms,
    random_crop,
    resize,
    rigid_transforms,
    to_tensor,
)


@pytest.mark.parametrize(
    "transform",
    [
        rigid_transforms,
        non_rigid_transforms,
        hue_saturation_transforms,
        blur_transforms,
        non_spatial_transforms,
    ],
)
def test_img_transforms(img_sample, transform):
    im = img_sample
    transforms = compose(transform())
    transformed = transforms(image=im)

    assert transformed["image"].shape == im.shape
    assert transformed["image"].dtype == im.dtype


@pytest.mark.parametrize(
    "transform",
    [center_crop, random_crop, resize],
)
def test_shape_transforms(img_sample, transform):
    expected_shape = (400, 420, 3)

    im = img_sample
    transforms = compose(transform(expected_shape[0], expected_shape[1]))
    transformed = transforms(image=im)

    assert transformed["image"].shape == expected_shape


def test_to_tensor(img_sample):
    transform = to_tensor()
    transformed = transform[0](image=img_sample)

    assert transformed["image"].dtype == torch.uint8


def test_tranform_pipeline(img_sample):
    trans_list = compose(
        [rigid_transforms(), non_rigid_transforms(), center_crop(450, 450), to_tensor()]
    )
    transformed = trans_list(image=img_sample)

    assert transformed["image"].shape == torch.Size([3, 450, 450])
    assert transformed["image"].dtype == torch.uint8



import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["Mish"]


@torch.jit.script
def mish_jit_fwd(x: torch.Tensor) -> torch.Tensor:
    """Mish forward."""
    return x.mul(torch.tanh(F.softplus(x)))


@torch.jit.script
def mish_jit_bwd(x: torch.Tensor, grad_output: torch.Tensor) -> torch.Tensor:
    """Mish backward."""
    x_sigmoid = torch.sigmoid(x)
    x_tanh_sp = F.softplus(x).tanh()

    return grad_output.mul(x_tanh_sp + x * x_sigmoid * (1 - x_tanh_sp * x_tanh_sp))


class MishFunction(torch.autograd.Function):
    """Memory efficient mish function implementation.

    Mish: A Self Regularized Non-Monotonic Neural Activation Function:

    https://arxiv.org/abs/1908.08681
    """

    @staticmethod
    def forward(ctx, x: torch.Tensor):
        """Forward pass."""
        ctx.save_for_backward(x)
        return mish_jit_fwd(x)

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        """Backward pass."""
        x = ctx.saved_tensors[0]
        return mish_jit_bwd(x, grad_output)


def mish(x: torch.Tensor):
    """Apply element-wise mish-activation.

    https://github.com/digantamisra98/Mish
    """
    return MishFunction.apply(x)


class Mish(nn.Module):
    def __init__(self, inplace: bool = False, **kwargs) -> None:
        """
        Element-wise mish.

        https://github.com/digantamisra98/Mish

        Parameters
        ----------
            inplace : bool, default=False
                This is not used, exists only for compatibility.
        """
        super().__init__()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass of the function.

        Parameters
        ----------
            input : torch.Tensor
                Input tensor. Can be of any shape (C, *).

        Returns
        -------
            torch.Tensor:
                activated output tensor. Shape same as input.
        """
        return mish(input)



from typing import List

import numpy as np

from cellseg_models_pytorch.utils import fix_duplicates

from ...utils import binarize
from ..functional import (
    gen_contour_maps,
    gen_dist_maps,
    gen_flow_maps,
    gen_hv_maps,
    gen_omni_flow_maps,
    gen_radial_distmaps,
    gen_stardist_maps,
    gen_weight_maps,
    smooth_distance,
)
from ._composition import OnlyInstMapTransform

__all__ = [
    "cellpose_transform",
    "hovernet_transform",
    "omnipose_transform",
    "stardist_transform",
    "stardist_opt_transform",
    "dist_transform",
    "smooth_dist_transform",
    "edgeweight_transform",
    "contour_transform",
    "binarize_transform",
]


class CellposeTrans(OnlyInstMapTransform):
    def __init__(self):
        """Generate flows from a heat diffused label mask.

        https://www.nature.com/articles/s41592-020-01018-x
        """
        super().__init__()
        self.name = "cellpose"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Fix duplicate values and generate flows.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Horizontal and vertical flows of objects.
                Shape: (2, H, W). Dtype: float64.
        """
        return gen_flow_maps(fix_duplicates(inst_map))


class HoVerNetTrans(OnlyInstMapTransform):
    def __init__(self):
        """Generate horizontal and vertical gradients from a label mask.

        https://www.sciencedirect.com/science/article/pii/S1361841519301045
        """
        super().__init__()
        self.name = "hovernet"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Fix duplicate values and generate gradients.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Horizontal and vertical gradients of objects.
                Shape: (2, H, W). Dtype: float64.
        """
        return gen_hv_maps(fix_duplicates(inst_map))


class OmniposeTrans(OnlyInstMapTransform):
    def __init__(self):
        """Generate horizontal and vertical eikonal flows from a label mask.

        https://www.biorxiv.org/content/10.1101/2021.11.03.467199v2
        """
        super().__init__()
        self.name = "omnipose"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Fix duplicate values and generate eikonal flows.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Horizontal and vertical gradients of objects.
                Shape: (2, H, W). Dtype: float64.
        """
        return gen_omni_flow_maps(fix_duplicates(inst_map))


# Not useful for now.
class StardistSelfTrans(OnlyInstMapTransform):
    def __init__(self, n_rays: int = 32, n_segments: int = 20):
        """Generate radial distance maps from a label mask.

        https://arxiv.org/abs/1806.03535

        Parameters
        ----------
            n_rays : int, default=32
                Number of rays used for computing distance maps.
            n_segments : int, default=20
                Number of line segments the contour is divided into.
                The more segments used, the more detail is preserved with
                performance tradeoff.
        """
        super().__init__()
        self.name = "stardist"
        self.n_rays = n_rays
        self.n_seg = n_segments

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Fix duplicate values and generate radial distance maps.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Pixelwise radial distance maps.
                Shape: (n_rays, H, W). Dtype: float64.
        """
        return gen_radial_distmaps(fix_duplicates(inst_map), self.n_rays, self.n_seg)


class StardistTrans(OnlyInstMapTransform):
    def __init__(self, n_rays: int = 32, **kwargs):
        """Generate radial distance maps from a label mask.

        https://arxiv.org/abs/1806.03535

        Parameters
        ----------
            n_rays : int, default=32
                Number of rays used for computing distance maps.
            n_segments : int, default=20
                Number of line segments the contour is divided into.
                The more segments used, the more detail is preserved with
                performance tradeoff.
        """
        super().__init__()
        self.name = "stardist"
        self.n_rays = n_rays

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Fix duplicate values and generate radial distance maps.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Pixelwise radial distance maps.
                Shape: (n_rays, H, W). Dtype: float64.
        """
        return gen_stardist_maps(fix_duplicates(inst_map), self.n_rays)


class SmoothDistTrans(OnlyInstMapTransform):
    def __init__(self):
        """Generate FIM distance transforms from a label mask.

        https://www.biorxiv.org/content/10.1101/2021.11.03.467199v2
        """
        super().__init__()
        self.name = "smoothdist"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Generate smooth distance transforms.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Smooth distance transforms of objects.
                Shape: (H, W). Dtype: float64.
        """
        return smooth_distance(inst_map)


class DistTrans(OnlyInstMapTransform):
    def __init__(self) -> None:
        """Generate distance transforms from a label mask."""
        super().__init__()
        self.name = "dist"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Generate distance transforms.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape: (H, W).

        Returns
        -------
            np.ndarray:
                Distance transforms of objects.
                Shape: (H, W). Dtype: float64.
        """
        return gen_dist_maps(fix_duplicates(inst_map))


class ContourTrans(OnlyInstMapTransform):
    def __init__(self):
        """Generate contour map from a label mask."""
        super().__init__()
        self.name = "contour"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Generate contour transforms.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Contour of objects. Shape: (H, W). Dtype: float64
        """
        return gen_contour_maps(fix_duplicates(inst_map))


class EdgeWeightTrans(OnlyInstMapTransform):
    def __init__(self):
        """Generate weight maps for object boundaries."""
        super().__init__()
        self.name = "edgeweight"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Generate edge weight transforms.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Contour of objects. Shape: (H, W). Dtype: float64
        """
        return gen_weight_maps(inst_map)


class BinarizeTrans(OnlyInstMapTransform):
    def __init__(self):
        """Binarize instance labelled mask."""
        super().__init__()
        self.name = "binary"

    def apply_to_instmap(self, inst_map: np.ndarray, **kwargs) -> np.ndarray:
        """Generate a binary mask from instance labelled mask.

        Parameters
        ----------
            inst_map : np.ndarray
                Instance labelled mask. Shape (H, W).

        Returns
        -------
            np.ndarray:
                Binary mask. Shape: (H, W). Dtype: uint8
        """
        return binarize(inst_map)


def cellpose_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the Cellpose tranformation for label mask."""
    return [CellposeTrans()]


def hovernet_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the HoVerNet tranformation for label mask."""
    return [HoVerNetTrans()]


def omnipose_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the omnipose tranformation for label mask."""
    return [OmniposeTrans()]


def stardist_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the stardist tranformation for label mask."""
    return [StardistTrans(**kwargs)]


def stardist_opt_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the stardist tranformation for label mask."""
    return [StardistSelfTrans(**kwargs)]


def dist_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the distance tranformation for label mask."""
    return [DistTrans()]


def smooth_dist_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the smooth distance tranformation for label mask."""
    return [SmoothDistTrans()]


def contour_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the contour tranformation for label mask."""
    return [ContourTrans()]


def edgeweight_transform(**kwargs) -> List[OnlyInstMapTransform]:
    """Return the edge weight tranformation for label mask."""
    return [EdgeWeightTrans()]


def binarize_transform(**kwargs):
    """Return the binarization tranformation for label mask."""
    return [BinarizeTrans()]
